#!/usr/bin/python
# -*- coding: utf-8 -*-
"""RT-DC dataset configuration"""
from __future__ import division, print_function, unicode_literals

import copy
import pathlib
import sys
import warnings


from ..compat import str_types
from .. import definitions as dfn


class UnknownConfigurationKeyWarning(UserWarning):
    pass


class ConfigurationDict(dict):
    def __init__(self, section=None, *args, **kwargs):
        """A case-insensitive dict that is section-aware

        Instantiate this dictionary like any other dictionary.
        All keys will be treated as lower-case keys. If `section`
        is given, new entries will be checked using the function
        :func:`verify_section_key`.
        """
        self.section = section
        super(ConfigurationDict, self).__init__(*args, **kwargs)
        self._convert_keys()

    def __getitem__(self, key):
        return super(ConfigurationDict,
                     self).__getitem__(self.__class__._k(key))

    def __setitem__(self, key, value):
        key = self.__class__._k(key)
        if self.section:
            valid = verify_section_key(self.section, key)
        else:
            valid = True
        if valid:
            # only set valid keys
            super(ConfigurationDict, self).__setitem__(key, value)

    def __delitem__(self, key):
        return super(ConfigurationDict,
                     self).__delitem__(self.__class__._k(key))

    def __contains__(self, key):
        return super(ConfigurationDict,
                     self).__contains__(self.__class__._k(key))

    @classmethod
    def _k(cls, key):
        """Convert a key to lower case"""
        return key.lower() if isinstance(key, str_types) else key

    def _convert_keys(self):
        for k in list(self.keys()):
            v = super(ConfigurationDict, self).pop(k)
            self.__setitem__(k, v)

    def get(self, key, *args, **kwargs):
        return super(ConfigurationDict,
                     self).get(self.__class__._k(key), *args, **kwargs)

    def items(self):
        keys = list(self.keys())
        keys.sort()
        out = [(k, self[k]) for k in keys]
        return out

    def pop(self, key, *args, **kwargs):
        return super(ConfigurationDict,
                     self).pop(self.__class__._k(key), *args, **kwargs)

    def setdefault(self, key, *args, **kwargs):
        return super(ConfigurationDict,
                     self).setdefault(self.__class__._k(key), *args, **kwargs)

    def update(self, E={}, **F):
        for key in E:
            self.__setitem__(key, E[key])
        for key in F:
            self.__setitem__(key, F[key])


class Configuration(object):
    def __init__(self, files=[], cfg={}, disable_checks=False):
        """Configuration class for RT-DC datasets

        This class has a dictionary-like interface to access
        and set configuration values, e.g.

        .. code::

            cfg = load_from_file("/path/to/config.txt")
            # access the channel width
            cfg["setup"]["channel width"]
            # modify the channel width
            cfg["setup"]["channel width"] = 30

        Parameters
        ----------
        files: list of files
            The config files with which to initialize the configuration
        cfg: dict-like
            The dictionary with which to initialize the configuration
        disable_checks: bool
            Set this to True if you want to avoid checking against
            section and key names defined in `dclab.definitions`
            using :func:`verify_section_key`. This avoids excess
            warning messages when loading data from configuration
            files not generated by dclab.
        """
        self.disable_checks = disable_checks

        self._cfg = ConfigurationDict()

        # set initial default values
        self._init_default_filter_values()

        # Update with additional dictionary
        self.update(cfg)

        # Load configuration files
        for f in files:
            self.update(load_from_file(f))

    def __contains__(self, key):
        return self._cfg.__contains__(key)

    def __getitem__(self, sec):
        if sec not in self and sec in dfn.config_keys:
            # create an empty section for user-convenience
            section = None if self.disable_checks else sec
            self._cfg[sec] = ConfigurationDict(section=section)
        item = self._cfg.__getitem__(sec)
        return item

    def __iter__(self):
        return self._cfg.__iter__()

    def __len__(self):
        return len(self._cfg)

    def __repr__(self):
        rep = ""
        keys = sorted(list(self.keys()))
        for key in keys:
            rep += "- {}\n".format(key)
            subkeys = sorted(list(self[key].keys()))
            for subkey in subkeys:
                rep += "   {}: {}\n".format(subkey, self[key][subkey])
        return rep

    def __setitem__(self, *args):
        self._cfg.__setitem__(*args)

    def _init_default_filter_values(self):
        """Set default initial values

        The default values are hard-coded for backwards compatibility
        and for several functionalities in dclab.
        """
        # Do not filter out invalid event values
        self["filtering"]["remove invalid events"] = False
        # Enable filters switch is mandatory
        self["filtering"]["enable filters"] = True
        # Limit events integer to downsample output data
        self["filtering"]["limit events"] = 0
        # Polygon filter list
        self["filtering"]["polygon filters"] = []
        # Defaults to no hierarchy parent
        self["filtering"]["hierarchy parent"] = "none"

        # Make sure that all filtering values have a default value
        # (otherwise we will get problems with resetting filters)
        for item in dfn.CFG_ANALYSIS["filtering"]:
            if item[0] not in self["filtering"]:
                raise KeyError(
                    "No default value set for [filtering]:{}".format(item[0]))

    def copy(self):
        """Return copy of current configuration"""
        return Configuration(cfg=copy.deepcopy(self._cfg))

    def keys(self):
        """Return the configuration keys (sections)"""
        return self._cfg.keys()

    def save(self, filename):
        """Save the configuration to a file"""
        filename = pathlib.Path(filename)
        out_str = self.tostring()
        with filename.open("w") as f:
            f.write(out_str)

    def tostring(self, sections=None):
        """Convert the configuration to its string representation

        The optional argument `sections` allows to export only
        specific sections of the configuration, i.e.
        `sections=dclab.dfn.CFG_METADATA` will only export
        configuration data from the original measurement and no
        filtering data.
        """
        out = []
        if sections is None:
            keys = sorted(list(self.keys()))
        else:
            keys = sorted([k for k in sections if k in self.keys()])
        for key in keys:
            out.append("[{}]".format(key))
            section = self[key]
            ikeys = list(section.keys())
            ikeys.sort()
            for ikey in ikeys:
                var, val = keyval_typ2str(ikey, section[ikey])
                out.append("{} = {}".format(var, val))
            out.append("")

        out_str = ""
        for i in range(len(out)):
            # win-like line endings
            out_str += out[i]+"\n"
        return out_str

    def update(self, newcfg):
        """Update current config with a dictionary"""
        for sec in newcfg.keys():
            if sec not in self._cfg:
                section = None if self.disable_checks else sec
                self._cfg[sec] = ConfigurationDict(section=section)
            self._cfg[sec].update(newcfg[sec])


class CaseInsensitiveDict(ConfigurationDict):
    def __init__(self, *args, **kwargs):
        warnings.warn("CaseInsensitiveDict is deprecated, use "
                      + "ConfigurationDict instead.",
                      DeprecationWarning)
        super(CaseInsensitiveDict, self).__init__(*args, **kwargs)


def verify_section_key(section, key):
    """Return True if the section-key combination exists"""
    wcount = 0
    if section in dfn.config_keys and key in dfn.config_keys[section]:
        pass
    elif section in ["plotting", "analysis"]:
        if sys.version_info[0] != 2:
            # hijacked by Shape-Out 1
            warnings.warn("The '{}' configuration key is deprecated!".format(
                section), UnknownConfigurationKeyWarning)
            wcount += 1
    elif section == "filtering":
        if key.endswith(" min") or key.endswith(" max"):
            feat = key[:-4]
            if not dfn.scalar_feature_exists(feat):
                warnings.warn(
                    "A range has been specified for an unknown feature "
                    + "'{}' in the 'filtering' section!".format(feat),
                    UnknownConfigurationKeyWarning)
                wcount += 1
        elif key == "limit events auto":
            # Shape-Out 1 used this to limit the number of events to
            # a common minimum.
            if sys.version_info[0] != 2:
                warnings.warn("The 'limit events auto' configuration key "
                              " in the 'filtering' section is deprecated!")
                wcount += 1
        else:
            warnings.warn(
                "Unknown key '{}' in the 'filtering' section!".format(key),
                UnknownConfigurationKeyWarning)
            wcount += 1
    elif section not in dfn.config_keys:
        warnings.warn("Unknown section '{}'!".format(section),
                      UnknownConfigurationKeyWarning)
        wcount += 1
    else:
        warnings.warn(
            "Unknown key '{}' in the '{}' section!".format(key, section),
            UnknownConfigurationKeyWarning)
        wcount += 1

    return wcount == 0


def load_from_file(cfg_file):
    """Load the configuration from a file

    Parameters
    ----------
    cfg_file: str
        Path to configuration file

    Returns
    -------
    cfg : ConfigurationDict
        Dictionary with configuration parameters
    """
    path = pathlib.Path(cfg_file).resolve()
    with path.open("r", errors="replace") as f:
        code = f.readlines()

    cfg = ConfigurationDict()
    for line in code:
        # We deal with comments and empty lines
        # We need to check line length first and then we look for
        # a hash.
        line = line.split("#")[0].strip()
        if len(line) != 0:
            if line.startswith("[") and line.endswith("]"):
                sec = line[1:-1].lower()
                if sec not in cfg:
                    cfg[sec] = ConfigurationDict()
                continue
            elif not line.count("="):
                # ignore invalid lines
                continue
            var, val = line.split("=", 1)
            var = var.strip().lower()
            val = val.strip("' ").strip('" ').strip()
            if len(val) == 0:
                # skip invalid values
                continue
            # convert parameter value to correct type
            if (sec in dfn.config_funcs and
                    var in dfn.config_funcs[sec]):
                # standard parameter with known type
                val = dfn.config_funcs[sec][var](val)
            else:
                # unknown parameter (e.g. plotting in Shape-Out), guess type
                var, val = keyval_str2typ(var, val)
            if len(var) != 0 and len(str(val)) != 0:
                cfg[sec][var] = val
    return cfg


def keyval_str2typ(var, val):
    """Convert a variable from a string to its correct type

    Parameters
    ----------
    var: str
        The variable name
    val: str
        The value of the variable represented as a string

    Returns
    -------
    varout: str
        Stripped lowercase `var`
    valout: any type
        The value converted from string to its presumed type

    Notes
    -----
    This method is heuristic and is only intended for usage in
    dclab.

    See Also
    --------
    keyval_typ2str: the opposite
    """
    if not (isinstance(val, str_types)):
        # already a type:
        return var.strip(), val
    var = var.strip().lower()
    val = val.strip()
    # Find values
    if len(var) != 0 and len(val) != 0:
        # check for float
        if val.startswith("[") and val.endswith("]"):
            if len(val.strip("[],")) == 0:
                # empty list
                values = []
            else:
                values = val.strip("[],").split(",")
                values = [float(v) for v in values]
            return var, values
        elif val.lower() in ["true", "y"]:
            return var, True
        elif val.lower() in ["false", "n"]:
            return var, False
        elif val[0] in ["'", '"'] and val[-1] in ["'", '"']:
            return var, val.strip("'").strip('"').strip()
        elif dfn.scalar_feature_exists(val):
            return var, val
        else:
            try:
                return var, float(val.replace(",", "."))
            except ValueError:
                return var, val


def keyval_typ2str(var, val):
    """Convert a variable to a string

    Parameters
    ----------
    var: str
        The variable name
    val: any type
        The value of the variable

    Returns
    -------
    varout: str
        Stripped lowercase `var`
    valout: any type
        The value converted to a useful string representation

    See Also
    --------
    keyval_str2typ: the opposite
    """
    varout = var.strip()
    if isinstance(val, list):
        data = ", ".join([keyval_typ2str(var, it)[1] for it in val])
        valout = "["+data+"]"
    elif isinstance(val, float):
        valout = "{:.12f}".format(val)
    else:
        valout = "{}".format(val)
    return varout, valout



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Testing script for volume.py
An ellipse is created and the analytical and numerical solution are compared
"""
from __future__ import division

import itertools as IT

import numpy as np

from dclab.features.volume import get_volume


def area_of_polygon(x, y):
    return 0.5 * np.abs(np.dot(x, np.roll(y, 1)) - np.dot(y, np.roll(x, 1)))


def centroid_of_polygon(points):
    """
    http://stackoverflow.com/a/14115494/190597 (mgamba)
    Centroid of polygon:
    http://en.wikipedia.org/wiki/Centroid#Centroid_of_polygon
    """
    area = area_of_polygon(*zip(*points))
    result_x = 0
    result_y = 0
    N = len(points)
    points = IT.cycle(points)
    x1, y1 = next(points)
    for _i in range(N):
        x0, y0 = x1, y1
        x1, y1 = next(points)
        cross = (x0 * y1) - (x1 * y0)
        result_x += (x0 + x1) * cross
        result_y += (y0 + y1) * cross
    result_x /= (area * 6.0)
    result_y /= (area * 6.0)
    return abs(result_x), abs(result_y)


def get_ellipse_coords(a, b, x=0.0, y=0.0, angle=0.0, k=2):
    """ Draws an ellipse using (360*k + 1) discrete points; based on pseudo code
    given at http://en.wikipedia.org/wiki/Ellipse
    k = 1 means 361 points (degree by degree)
    a = major axis distance,
    b = minor axis distance,
    x = offset along the x-axis
    y = offset along the y-axis
    angle = clockwise rotation [in degrees] of the ellipse;
        * angle=0  : the ellipse is aligned with the positive x-axis
        * angle=30 : rotated 30 degrees clockwise from positive x-axis
    """
    pts = np.zeros((360 * k + 1, 2))

    beta = -angle * np.pi / 180
    sin_beta = np.sin(beta)
    cos_beta = np.cos(beta)
    alpha = np.radians(np.r_[0.:360.:1j * (360 * k + 1)])

    sin_alpha = np.sin(alpha)
    cos_alpha = np.cos(alpha)

    pts[:, 0] = x + (a * cos_alpha * cos_beta - b * sin_alpha * sin_beta)
    pts[:, 1] = y + (a * cos_alpha * sin_beta + b * sin_alpha * cos_beta)

    return pts


def test_orientation():
    """counter-clockwise vs clockwise"""
    # Helper definitions to get an ellipse
    major = 10
    minor = 5
    ellip = get_ellipse_coords(a=major,
                               b=minor,
                               x=minor,
                               y=5,
                               angle=0,
                               k=100)
    # obtain the centroid (corresponds to pos_x and pos_lat)
    cx, cy = centroid_of_polygon(ellip)

    v1 = get_volume(cont=ellip,
                    pos_x=cx,
                    pos_y=cy,
                    pix=1)

    # Turn contour around
    v2 = get_volume(cont=ellip[::-1, :],
                    pos_x=cx,
                    pos_y=cy,
                    pix=1)

    assert np.all(v1 == v2)


def test_get_volume():
    # Helper definitions to get an ellipse
    major = 10
    minor = 5
    ellip = get_ellipse_coords(a=major,
                               b=minor,
                               x=minor,
                               y=5,
                               angle=0,
                               k=100)
    # obtain the centroid (corresponds to pos_x and pos_lat)
    cx, cy = centroid_of_polygon(ellip)
    elliplist = []
    elliplist.append(ellip)
    volume = get_volume(cont=elliplist,
                        pos_x=[cx],
                        pos_y=[cy],
                        pix=1)

    # Analytic solution for volume of ellipsoid:
    V = 4 / 3 * np.pi * major * minor**2
    msg = "Calculation of volume is faulty!"
    assert np.allclose(np.array(volume), np.array(V)), msg


def test_shape():
    major = 10
    minor = 5
    ellip = get_ellipse_coords(a=major,
                               b=minor,
                               x=minor,
                               y=5.0,
                               angle=0,
                               k=100)
    cx, cy = centroid_of_polygon(ellip)
    # no lists
    volume = get_volume(cont=ellip,
                        pos_x=cx,
                        pos_y=cy,
                        pix=1)
    assert isinstance(volume, float)

    volumelist = get_volume(cont=[ellip],
                            pos_x=[cx],
                            pos_y=[cy],
                            pix=1)
    assert isinstance(volumelist, np.ndarray)


def test_xpos():
    """xpos is not necessary to compute volume dense ellipse"""
    major = 10
    minor = 5
    ellip = get_ellipse_coords(a=major,
                               b=minor,
                               x=minor,
                               y=5.0,
                               angle=0,
                               k=100)
    cx, cy = centroid_of_polygon(ellip)
    # no lists
    v0 = get_volume(cont=ellip,
                    pos_x=cx,
                    pos_y=cy,
                    pix=1)
    for cxi in np.linspace(0, 2 * cx, 10):
        vi = get_volume(cont=ellip,
                        pos_x=cxi,
                        pos_y=cy,
                        pix=1)
        assert np.allclose(v0, vi)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import division

import pathlib
import warnings

import numpy as np

import dclab
from dclab import isoelastics as iso
from dclab.features import emodulus
from dclab.features.emodulus import pxcorr

from helper_methods import example_data_dict


def get_isofile(name="example_isoelastics.txt"):
    thisdir = pathlib.Path(__file__).parent
    return thisdir / "data" / name


def test_bad_isoelastic():
    i1 = iso.Isoelastics([get_isofile()])
    try:
        i1.get(col1="deform",
               col2="area_ratio",
               method="analytical",
               channel_width=20,
               flow_rate=0.04,
               viscosity=15,
               add_px_err=False,
               px_um=None)
    except KeyError:
        pass
    else:
        assert False, "features should not work"


def test_bad_isoelastic_2():
    i1 = iso.Isoelastics([get_isofile()])
    try:
        i1.get(col1="deform",
               col2="area_um",
               method="numerical",
               channel_width=20,
               flow_rate=0.04,
               viscosity=15,
               add_px_err=False,
               px_um=None)
    except KeyError:
        pass
    else:
        assert False, "only analytical should not work with this set"


def test_bad_isoelastic_3():
    i1 = iso.Isoelastics([get_isofile()])
    try:
        i1.get(col1="deform",
               col2="bad_feature",
               method="numerical",
               channel_width=20,
               flow_rate=0.04,
               viscosity=15,
               add_px_err=False,
               px_um=None)
    except ValueError:
        pass
    else:
        assert False, "bad feature does not work"


def test_bad_isoelastic_4():
    i1 = iso.Isoelastics([get_isofile()])
    try:
        i1.get(col1="deform",
               col2="area_um",
               method="monsterelastic",
               channel_width=20,
               flow_rate=0.04,
               viscosity=15,
               add_px_err=False,
               px_um=None)
    except ValueError:
        pass
    else:
        assert False, "bad method does not work"


def test_circ():
    i1 = iso.Isoelastics([get_isofile()])
    iso1 = i1._data["analytical"]["area_um"]["deform"]["isoelastics"]
    iso2 = i1._data["analytical"]["area_um"]["circ"]["isoelastics"]
    assert np.allclose(iso1[0][:, 1], 1 - iso2[0][:, 1])


def test_circ_get():
    i1 = iso.Isoelastics([get_isofile()])
    iso_circ = i1.get(col1="area_um",
                      col2="circ",
                      method="analytical",
                      channel_width=15,
                      flow_rate=0.04,
                      viscosity=15)
    iso_deform = i1.get(col1="area_um",
                        col2="deform",
                        method="analytical",
                        channel_width=15,
                        flow_rate=0.04,
                        viscosity=15)
    for ii in range(len(iso_circ)):
        isc = iso_circ[ii]
        isd = iso_deform[ii]
        assert np.allclose(isc[:, 0], isd[:, 0])
        assert np.allclose(isc[:, 1], 1 - isd[:, 1])


def test_convert():
    i1 = iso.Isoelastics([get_isofile()])
    isoel = i1._data["analytical"]["area_um"]["deform"]["isoelastics"]
    isoel15 = i1.convert(isoel=isoel,
                         col1="area_um",
                         col2="deform",
                         channel_width_in=20,
                         channel_width_out=15,
                         flow_rate_in=0.04,
                         flow_rate_out=0.04,
                         viscosity_in=15,
                         viscosity_out=15)
    # These values were taken from previous isoelasticity files
    # used in Shape-Out.
    assert np.allclose(isoel15[0][:, 2], 7.11111111e-01)
    assert np.allclose(isoel15[1][:, 2], 9.48148148e-01)
    # area_um
    assert np.allclose(isoel15[0][1, 0], 2.245995843750000276e+00)
    assert np.allclose(isoel15[0][9, 0], 9.954733499999999680e+00)
    assert np.allclose(isoel15[1][1, 0], 2.247747243750000123e+00)
    # deform
    assert np.allclose(isoel15[0][1, 1], 5.164055600000000065e-03)
    assert np.allclose(isoel15[0][9, 1], 2.311524599999999902e-02)
    assert np.allclose(isoel15[1][1, 1], 2.904264599999999922e-03)


def test_convert_error():
    i1 = iso.Isoelastics([get_isofile()])
    isoel = i1.get(col1="area_um",
                   col2="deform",
                   method="analytical",
                   channel_width=15)

    kwargs = dict(channel_width_in=15,
                  channel_width_out=20,
                  flow_rate_in=.12,
                  flow_rate_out=.08,
                  viscosity_in=15,
                  viscosity_out=15)

    try:
        i1.convert(isoel=isoel,
                   col1="deform",
                   col2="area_ratio",
                   **kwargs)
    except KeyError:
        pass
    except BaseException:
        raise
    else:
        assert False, "undefined column volume"


def test_data_slicing():
    i1 = iso.Isoelastics([get_isofile()])
    iso1 = i1._data["analytical"]["area_um"]["deform"]["isoelastics"]
    iso2 = i1._data["analytical"]["deform"]["area_um"]["isoelastics"]
    for ii in range(len(iso1)):
        assert np.all(iso1[ii][:, 2] == iso2[ii][:, 2])
        assert np.all(iso1[ii][:, 0] == iso2[ii][:, 1])
        assert np.all(iso1[ii][:, 1] == iso2[ii][:, 0])


def test_data_structure():
    i1 = iso.Isoelastics([get_isofile()])
    # basic import
    assert "analytical" in i1._data
    assert "deform" in i1._data["analytical"]
    assert "area_um" in i1._data["analytical"]["deform"]
    assert "area_um" in i1._data["analytical"]
    assert "deform" in i1._data["analytical"]["area_um"]
    # circularity
    assert "circ" in i1._data["analytical"]
    assert "area_um" in i1._data["analytical"]["circ"]
    assert "area_um" in i1._data["analytical"]
    assert "circ" in i1._data["analytical"]["area_um"]
    # metadata
    meta1 = i1._data["analytical"]["area_um"]["deform"]["meta"]
    meta2 = i1._data["analytical"]["deform"]["area_um"]["meta"]
    assert meta1 == meta2


def test_get():
    i1 = iso.Isoelastics([get_isofile()])
    data = i1.get(col1="area_um",
                  col2="deform",
                  channel_width=20,
                  flow_rate=0.04,
                  viscosity=15,
                  method="analytical")
    refd = i1._data["analytical"]["area_um"]["deform"]["isoelastics"]

    for a, b in zip(data, refd):
        assert np.all(a == b)


def test_pixel_err():
    i1 = iso.Isoelastics([get_isofile()])
    isoel = i1._data["analytical"]["area_um"]["deform"]["isoelastics"]
    px_um = .10
    # add the error
    isoel_err = i1.add_px_err(isoel=isoel,
                              col1="area_um",
                              col2="deform",
                              px_um=px_um,
                              inplace=False)
    # remove the error manually
    isoel_corr = []
    for iss in isoel_err:
        iss = iss.copy()
        iss[:, 1] -= pxcorr.corr_deform_with_area_um(area_um=iss[:, 0],
                                                     px_um=px_um)
        isoel_corr.append(iss)

    for ii in range(len(isoel)):
        assert not np.allclose(isoel[ii], isoel_err[ii])
        assert np.allclose(isoel[ii], isoel_corr[ii])

    try:
        i1.add_px_err(isoel=isoel,
                      col1="deform",
                      col2="deform",
                      px_um=px_um,
                      inplace=False)
    except ValueError:
        pass
    else:
        assert False, "identical columns"

    try:
        i1.add_px_err(isoel=isoel,
                      col1="deform",
                      col2="circ",
                      px_um=px_um,
                      inplace=False)
    except KeyError:
        pass
    except BaseException:
        raise
    else:
        assert False, "area_um required"


def test_volume_basic():
    """Reproduce exact data from simulation result"""
    i1 = iso.get_default()
    data = i1.get(col1="volume",
                  col2="deform",
                  channel_width=20,
                  flow_rate=0.04,
                  viscosity=15,
                  method="numerical",
                  add_px_err=False,
                  px_um=None)
    assert np.allclose(data[0][0], [1.61819e+02, 4.18005e-02, 1.08000e+00])
    assert np.allclose(data[0][-1], [5.90127e+02, 1.47449e-01, 1.08000e+00])
    assert np.allclose(data[1][0], [1.61819e+02, 2.52114e-02, 1.36000e+00])
    assert np.allclose(data[-1][-1], [3.16212e+03, 1.26408e-02, 1.08400e+01])


def test_volume_pxcorr():
    """Deformation is pixelation-corrected using volume"""
    i1 = iso.get_default()
    data = i1.get(col1="volume",
                  col2="deform",
                  channel_width=20,
                  flow_rate=None,
                  viscosity=None,
                  method="numerical",
                  add_px_err=True,
                  px_um=0.34)
    ddelt = pxcorr.corr_deform_with_volume(1.61819e+02, px_um=0.34)
    assert np.allclose(data[0][0], [1.61819e+02,
                                    4.18005e-02 + ddelt,
                                    1.08000e+00])


def test_volume_scale():
    """Simple volume scale"""
    i1 = iso.get_default()
    data = i1.get(col1="volume",
                  col2="deform",
                  channel_width=25,
                  flow_rate=0.04,
                  viscosity=15,
                  method="numerical",
                  add_px_err=False,
                  px_um=None)

    assert np.allclose(data[0][0], [1.61819e+02 * (25 / 20)**3,
                                    4.18005e-02,
                                    1.08000e+00 * (20 / 25)**3])


def test_volume_scale_2():
    """The default values are used if set to None"""
    i1 = iso.get_default()
    data = i1.get(col1="volume",
                  col2="deform",
                  channel_width=25,
                  flow_rate=None,
                  viscosity=None,
                  method="numerical",
                  add_px_err=False,
                  px_um=None)
    assert np.allclose(data[0][0], [1.61819e+02 * (25 / 20)**3,
                                    4.18005e-02,
                                    1.08000e+00 * (20 / 25)**3])


def test_volume_switch():
    """Switch the columns"""
    i1 = iso.get_default()
    data = i1.get(col1="deform",
                  col2="volume",
                  channel_width=20,
                  flow_rate=0.04,
                  viscosity=15,
                  method="numerical",
                  add_px_err=False,
                  px_um=None)
    assert np.allclose(data[0][0], [4.18005e-02, 1.61819e+02, 1.08000e+00])
    assert np.allclose(data[-1][-1], [1.26408e-02, 3.16212e+03, 1.08400e+01])


def test_volume_switch_scale():
    """Switch the columns and change the scale"""
    i1 = iso.get_default()
    data = i1.get(col1="deform",
                  col2="volume",
                  channel_width=25,
                  flow_rate=0.04,
                  viscosity=15,
                  method="numerical",
                  add_px_err=False,
                  px_um=None)
    assert np.allclose(data[0][0], [4.18005e-02,
                                    1.61819e+02 * (25 / 20)**3,
                                    1.08000e+00 * (20 / 25)**3])
    assert np.allclose(data[-1][-1], [1.26408e-02,
                                      3.16212e+03 * (25 / 20)**3,
                                      1.08400e+01 * (20 / 25)**3])


def test_with_rtdc():
    keys = ["area_um", "deform"]
    ddict = example_data_dict(size=8472, keys=keys)
    # legacy
    ds = dclab.new_dataset(ddict)
    ds.config["setup"]["flow rate"] = 0.16
    ds.config["setup"]["channel width"] = 30
    ds.config["setup"]["temperature"] = 23.0
    ds.config["setup"]["medium"] = "CellCarrier"
    ds.config["imaging"]["pixel size"] = .34
    i1 = iso.get_default()
    data1 = i1.get_with_rtdcbase(col1="area_um",
                                 col2="deform",
                                 method="numerical",
                                 dataset=ds,
                                 viscosity=None,
                                 add_px_err=False)

    viscosity = emodulus.viscosity.get_viscosity(
        medium="CellCarrier",
        channel_width=ds.config["setup"]["channel width"],
        flow_rate=ds.config["setup"]["flow rate"],
        temperature=ds.config["setup"]["temperature"])
    data2 = i1.get(col1="area_um",
                   col2="deform",
                   method="numerical",
                   channel_width=ds.config["setup"]["channel width"],
                   flow_rate=ds.config["setup"]["flow rate"],
                   viscosity=viscosity,
                   add_px_err=False,
                   px_um=ds.config["imaging"]["pixel size"])
    for d1, d2 in zip(data1, data2):
        assert np.allclose(d1, d2, atol=0, rtol=1e-14)


def test_with_rtdc_warning():
    keys = ["area_um", "deform"]
    ddict = example_data_dict(size=8472, keys=keys)
    # legacy
    ds = dclab.new_dataset(ddict)
    ds.config["setup"]["flow rate"] = 0.16
    ds.config["setup"]["channel width"] = 30
    ds.config["setup"]["medium"] = "CellCarrier"
    ds.config["imaging"]["pixel size"] = .34
    i1 = iso.get_default()
    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        # Trigger a warning (temperature missing).
        i1.get_with_rtdcbase(col1="area_um",
                             col2="deform",
                             method="numerical",
                             dataset=ds,
                             viscosity=None,
                             add_px_err=False)
        # Verify some things
        assert len(w) == 1
        assert issubclass(w[-1].category,
                          iso.IsoelasticsEmodulusMeaninglessWarning)
        assert "plotting" in str(w[-1].message)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""Handling fluorescence trace data"""
from __future__ import division, print_function, unicode_literals

import pathlib
import warnings

from nptdms import TdmsFile
import numpy as np

from ... import definitions as dfn

from . import naming
from .exc import InvalidTDMSFileFormatError


class MultipleSamplesPerEventFound(UserWarning):
    pass


class TraceColumn(object):
    def __init__(self, rtdc_dataset):
        """Prepares everything but does not load the trace data yet

        The trace data is loaded when __getitem__, __len__, or __iter__
        are called. This saves time and memory when the trace data is
        not needed at all, e.g. for batch processing with Shape-Out.
        """
        self._trace = None
        self.mname = rtdc_dataset.path
        self.identifier = self.mname

    def __getitem__(self, trace_key):
        if trace_key not in dfn.FLUOR_TRACES:
            msg = "Unknown fluorescence trace key: {}".format(trace_key)
            raise ValueError(msg)
        return self.trace.__getitem__(trace_key)

    def __len__(self):
        return self.trace.__len__()

    def __iter__(self):
        return self.trace.__iter__()

    def __repr__(self):
        tname = TraceColumn.find_trace_file(self.mname)
        if self._trace is None:
            addstr = "not loaded into memory"
        else:
            addstr = "loaded into memory"

        if tname is None:
            rep = "No trace data available!"
        else:
            rep = "Fluorescence trace data from file {}, <{}>".format(tname,
                                                                      addstr)
        return rep

    def keys(self):
        return self.trace.keys()

    @property
    def trace(self):
        """Initializes the trace data"""
        if self._trace is None:
            self._trace = self.load_trace(self.mname)
        return self._trace

    @staticmethod
    def load_trace(mname):
        """Loads the traces and returns them as a dictionary

        Currently, only loading traces from tdms files is supported.
        This forces us to load the full tdms file into memory which
        takes some time.
        """
        tname = TraceColumn.find_trace_file(mname)

        # Initialize empty trace dictionary
        trace = {}

        if tname is None:
            pass
        elif tname.suffix == ".tdms":
            # Again load the measurement tdms file.
            # This might increase memory usage, but it is cleaner
            # when looking at code structure.
            mdata = TdmsFile(str(mname))
            try:
                sampleids = mdata["Cell Track"]["FL1index"].data
            except KeyError:
                raise InvalidTDMSFileFormatError(
                    "No 'FL1index' column in '{}'!".format(tname))

            # Check that sample IDs are always incremented with same
            # sample size.
            samples_per_event = np.unique(np.diff(sampleids))
            if len(samples_per_event) > 1:
                # This means the length of the fluorescence trace is not
                # a constant. According to Philipp, this means the trace
                # cannot be used.
                warnings.warn("Ignoring trace data of '{}' ".format(tname)
                              + "due to multiple values for samples per "
                              + "event: {}".format(samples_per_event),
                              MultipleSamplesPerEventFound)
            else:
                # Load the trace data. The traces file is usually larger than
                # the measurement file.
                tdata = TdmsFile(str(tname))
                for trace_key in dfn.FLUOR_TRACES:
                    group, ch = naming.tr_data_map[trace_key]
                    try:
                        trdat = tdata[group][ch].data
                    except KeyError:
                        pass
                    else:
                        if trdat is not None and trdat.size != 0:
                            # Split the input trace data into equally-saced
                            # sections (we already tested that sampleids is
                            # equally-spaced).
                            spe = sampleids[1] - sampleids[0]
                            trace[trace_key] = np.split(trdat, trdat.size//spe)
        return trace

    @staticmethod
    def find_trace_file(mname):
        """Tries to find the traces tdms file name

        Returns None if no trace file is found.
        """
        mname = pathlib.Path(mname)
        tname = None

        if mname.exists():
            cand = mname.with_name(mname.name[:-5] + "_traces.tdms")
            if cand.exists():
                tname = cand

        return tname



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" Test filter hierarchies
"""
from __future__ import print_function

import numpy as np
import pytest

from dclab import new_dataset
from dclab.rtdc_dataset import fmt_hierarchy

from helper_methods import example_data_dict, example_data_sets, cleanup, \
    retrieve_data


def test_event_count():
    tdms_path = retrieve_data(example_data_sets[1])
    ds = new_dataset(tdms_path)
    ds.filter.manual[0] = False
    ch = new_dataset(ds)
    assert ds.config["experiment"]["event count"] == len(ds)
    assert ch.config["experiment"]["event count"] == len(ch)
    assert len(ds) == len(ch) + 1
    cleanup()


@pytest.mark.filterwarnings(
    'ignore::dclab.rtdc_dataset.config.UnknownConfigurationKeyWarning')
def test_feat_contour():
    path = retrieve_data("rtdc_data_hdf5_contour_image_trace.zip")
    ds = new_dataset(path)
    ds.filter.manual[0] = False
    ds.filter.manual[2] = False
    ch = new_dataset(ds)
    assert np.all(ch["contour"][0] == ds["contour"][1])
    assert np.all(ch["contour"][1] == ds["contour"][3])
    cleanup()


@pytest.mark.filterwarnings(
    'ignore::dclab.rtdc_dataset.config.UnknownConfigurationKeyWarning')
def test_feat_image():
    path = retrieve_data("rtdc_data_hdf5_contour_image_trace.zip")
    ds = new_dataset(path)
    ds.filter.manual[0] = False
    ds.filter.manual[2] = False
    ch = new_dataset(ds)
    assert np.all(ch["image"][0] == ds["image"][1])
    assert np.all(ch["image"][1] == ds["image"][3])
    cleanup()


@pytest.mark.filterwarnings(
    'ignore::dclab.rtdc_dataset.config.UnknownConfigurationKeyWarning')
def test_feat_mask():
    path = retrieve_data("rtdc_data_hdf5_mask_contour.zip")
    ds = new_dataset(path)
    ds.filter.manual[0] = False
    ds.filter.manual[2] = False
    ch = new_dataset(ds)
    assert np.all(ch["mask"][0] == ds["mask"][1])
    assert np.all(ch["mask"][1] == ds["mask"][3])
    cleanup()


@pytest.mark.filterwarnings(
    'ignore::dclab.rtdc_dataset.config.UnknownConfigurationKeyWarning')
def test_feat_trace():
    path = retrieve_data("rtdc_data_hdf5_contour_image_trace.zip")
    ds = new_dataset(path)
    ds.filter.manual[0] = False
    ds.filter.manual[2] = False
    ch = new_dataset(ds)
    assert np.all(ch["trace"]["fl1_median"][0]
                  == ds["trace"]["fl1_median"][1])
    assert np.all(ch["trace"]["fl1_median"][1]
                  == ds["trace"]["fl1_median"][3])
    cleanup()


@pytest.mark.filterwarnings(
    'ignore::dclab.rtdc_dataset.config.UnknownConfigurationKeyWarning')
def test_features():
    path = retrieve_data("rtdc_data_hdf5_contour_image_trace.zip")
    ds = new_dataset(path)
    ch = new_dataset(ds)
    assert ds.features == ch.features
    assert ds.features_innate == ch.features_innate
    assert ds.features_loaded == ch.features_loaded
    assert ds.features_scalar == ch.features_scalar
    cleanup()


@pytest.mark.filterwarnings(
    'ignore::dclab.rtdc_dataset.config.UnknownConfigurationKeyWarning')
def test_features_loaded():
    path = retrieve_data("rtdc_data_hdf5_contour_image_trace.zip")
    ds = new_dataset(path)
    ch = new_dataset(ds)
    assert "volume" in ch.features
    assert "volume" not in ch.features_loaded
    assert "volume" not in ch.features_innate
    # compute volume, now it should be loaded
    ch["volume"]
    assert "volume" in ch.features_loaded
    assert "volume" in ds.features_loaded
    cleanup()


def test_hierarchy_from_tdms():
    tdms_path = retrieve_data(example_data_sets[1])
    ds1 = new_dataset(tdms_path)
    ds2 = new_dataset(ds1)

    ds1.filter.manual[0] = False
    ds2.apply_filter()
    assert ds2.filter.all.shape[0] == ds1.filter.all.shape[0] - 1
    assert ds2["area_um"][0] == ds1["area_um"][1]
    cleanup()


def test_index_deep_contour():
    data = example_data_dict(42, keys=["area_um", "contour", "deform"])
    ds = new_dataset(data)
    ds.filter.manual[3] = False
    c1 = new_dataset(ds)
    c1.filter.manual[1] = False
    c2 = new_dataset(c1)
    assert np.all(c2["contour"][3] == ds["contour"][5])


def test_manual_exclude():
    data = example_data_dict(42, keys=["area_um", "deform"])
    p = new_dataset(data)
    c1 = new_dataset(p)
    c2 = new_dataset(c1)
    c3 = new_dataset(c2)
    c1.filter.manual[0] = False
    c2.apply_filter()
    c2.filter.manual[1] = False
    c3.apply_filter()

    # simple exclusion of few events
    assert len(c3) == len(p) - 2

    # removing same event in parent removes the event from the
    # child altogether, including the manual filter
    c3.filter.manual[0] = False
    c2.filter.manual[0] = False
    c3.apply_filter()
    assert np.alltrue(c3.filter.manual)

    # reinserting the event in the parent, retrieves back
    # the manual filter in the child
    c2.filter.manual[0] = True
    c3.apply_filter()
    assert not c3.filter.manual[0]


def test_manual_exclude_parent_changed():
    data = example_data_dict(42, keys=["area_um", "tilt"])
    p = new_dataset(data)
    p.filter.manual[4] = False
    c = new_dataset(p)
    c.filter.manual[5] = False
    c.apply_filter()
    p.config["filtering"]["tilt min"] = 0
    p.config["filtering"]["tilt max"] = .5
    p.apply_filter()
    assert np.sum(p.filter.all) == 21
    # size of child is directly determined from parent
    assert len(c) == 21
    # filters have not yet been updated
    assert len(c.filter.all) == 41
    assert c.filter.parent_changed
    # the initially excluded event
    assert c.filter.retrieve_manual_indices(c) == [6]

    # try to change the excluded events
    try:
        c.filter.apply_manual_indices(c, [1, 2])
    except fmt_hierarchy.HierarchyFilterError:
        pass
    else:
        assert False, "expected HierarchyFilterError"

    # this can be resolved by applying the filter
    c.apply_filter()
    c.filter.apply_manual_indices(c, [1, 2])
    assert c.filter.retrieve_manual_indices(c) == [1, 2]


def test_same_hash_different_identifier():
    tdms_path = retrieve_data(example_data_sets[1])
    ds1 = new_dataset(tdms_path)
    ds1.filter.manual[0] = False
    ch1 = new_dataset(ds1)
    ch2 = new_dataset(ds1)
    assert len(ch1) == len(ds1) - 1
    assert ch1.hash == ch2.hash
    assert ch1.identifier != ch2.identifier
    cleanup()


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



#!/usr/bin/python
# -*- coding: utf-8 -*-
from __future__ import division, print_function, unicode_literals

import numpy as np
from .ancillary_feature import AncillaryFeature


def get_ml_score_names(mm):
    """Return a list of all ml_score_??? features"""
    feats = []
    # We cannot loop over mm.features because of infinite recursions
    for ft in mm._feature_candidates:
        if ft.startswith("ml_score_") and ft in mm:
            feats.append(ft)
    return sorted(feats)


def compute_ml_class(mm):
    """Compute the most-probable class membership for all events

    This method also checks whether

    - there are any nan values in the ml_score features
    - the ml_score features are in the interval [0,1]

    Notes
    -----
    I initially thought about also checking whether each feature
    sums to one, but discarded the idea. Assume that a classifier
    does a really bad classification and classifies all events in
    the same way. If the dataset is cropped at some point (e.g.
    debris or other events), then this bad classifier has an
    increased probability compared to another classifier which is
    really good at picking out one population. The ml_score values
    should be just in the range of [0, 1]. This also simplifies
    export to hdf5 and the work with hierarchy children.
    """
    feats = get_ml_score_names(mm)
    # first check each feature for nans and boundaries
    for ft in feats:
        if np.sum(np.isnan(mm[ft])):
            raise ValueError("Feature '{}' has nan values!".format(ft))
        elif np.sum(mm[ft] > 1):
            raise ValueError("Feature '{}' has values > 1!".format(ft))
        elif np.sum(mm[ft] < 0):
            raise ValueError("Feature '{}' has values < 0!".format(ft))
    # now compute the maximum for each event
    scores = [mm[ft] for ft in feats]
    matrix = np.vstack(scores)
    ml_class = np.argmax(matrix, axis=0)
    return ml_class


def has_ml_scores(mm):
    """Check whether the dataset has ml_scores defined"""
    return bool(get_ml_score_names(mm))


def register():
    AncillaryFeature(feature_name="ml_class",
                     method=compute_ml_class,
                     req_func=has_ml_scores,
                     )



#!/usr/bin/python
# -*- coding: utf-8 -*-
from __future__ import division, print_function, unicode_literals

from .compat import str_types


def fbool(value):
    """boolean"""
    if isinstance(value, str_types):
        value = value.lower()
        if value == "false":
            value = False
        elif value == "true":
            value = True
        elif value:
            value = bool(float(value))
        else:
            raise ValueError("empty string")
    else:
        value = bool(float(value))
    return value


def fint(value):
    """integer"""
    if isinstance(value, str_types):
        # strings might have been saved wrongly as booleans
        value = value.lower()
        if value == "false":
            value = 0
        elif value == "true":
            value = 1
        elif value:
            value = int(float(value))
        else:
            raise ValueError("empty string")
    else:
        value = int(float(value))
    return value


def fintlist(alist):
    """A list of integers"""
    outlist = []
    if not isinstance(alist, (list, tuple)):
        # we have a string (comma-separated integers)
        alist = alist.strip().strip("[] ").split(",")
    for it in alist:
        if it:
            outlist.append(fint(it))
    return outlist


def lcstr(astr):
    """lower-case string"""
    return astr.lower()


#: maps functions to their expected output types
func_types = {fbool: bool,
              fint: int,
              fintlist: list,
              lcstr: str_types}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function

import time
import sys

import pytest

from dclab import cached


@pytest.mark.skipif(sys.version_info < (3, 3),
                    reason="perf_counter requires python3.3 or higher")
def test_cache():
    "Test if caching happens"
    wait = .05

    @cached.Cache
    def func1(x):
        time.sleep(wait)
        return 2 * x

    a = func1(4)
    assert a == 8
    t1 = time.perf_counter()
    b = func1(4)
    t2 = time.perf_counter()
    assert t2 - t1 < wait
    assert b == a


@pytest.mark.skipif(sys.version_info < (3, 3),
                    reason="perf_counter requires python3.3 or higher")
def test_cache_size():
    "Create more caches than cached.MAX_SIZE"
    ss = 10
    wait = .01
    cached.MAX_SIZE = ss

    @cached.Cache
    def func1(x):
        time.sleep(wait)
        return 2 * x
    for ii in range(ss):
        func1(ii)

    t1 = time.perf_counter()
    for ii in range(ss):
        func1(ii)
    t2 = time.perf_counter()
    assert t2 - t1 < wait

    func1(3.14)
    t3 = time.perf_counter()
    assert t3 - t2 > wait


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



"""
Multivariate Conditional and Unconditional Kernel Density Estimation
with Mixed Data Types.

References
----------
[1] Racine, J., Li, Q. Nonparametric econometrics: theory and practice.
    Princeton University Press. (2007)
[2] Racine, Jeff. "Nonparametric Econometrics: A Primer," Foundation
    and Trends in Econometrics: Vol 3: No 1, pp1-88. (2008)
    http://dx.doi.org/10.1561/0800000009
[3] Racine, J., Li, Q. "Nonparametric Estimation of Distributions
    with Categorical and Continuous Data." Working Paper. (2000)
[4] Racine, J. Li, Q. "Kernel Estimation of Multivariate Conditional
    Distributions Annals of Economics and Finance 5, 211-235 (2004)
[5] Liu, R., Yang, L. "Kernel estimation of multivariate
    cumulative distribution function."
    Journal of Nonparametric Statistics (2008)
[6] Li, R., Ju, G. "Nonparametric Estimation of Multivariate CDF
    with Categorical and Continuous Data." Working Paper
[7] Li, Q., Racine, J. "Cross-validated local linear nonparametric
    regression" Statistica Sinica 14(2004), pp. 485-512
[8] Racine, J.: "Consistent Significance Testing for Nonparametric
        Regression" Journal of Business & Economics Statistics
[9] Racine, J., Hart, J., Li, Q., "Testing the Significance of
        Categorical Predictor Variables in Nonparametric Regression
        Models", 2006, Econometric Reviews 25, 523-544

"""
from __future__ import division

import numpy as np

from ._kernel_base import GenericKDE, EstimatorSettings, gpke, _adjust_shape


__all__ = ['KDEMultivariate', 'EstimatorSettings']


class KDEMultivariate(GenericKDE):
    """
    Multivariate kernel density estimator.

    This density estimator can handle univariate as well as multivariate data,
    including mixed continuous / ordered discrete / unordered discrete data.
    It also provides cross-validated bandwidth selection methods (least
    squares, maximum likelihood).

    Parameters
    ----------
    data: list of ndarrays or 2-D ndarray
        The training data for the Kernel Density Estimation, used to determine
        the bandwidth(s).  If a 2-D array, should be of shape
        (num_observations, num_variables).  If a list, each list element is a
        separate observation.
    var_type: str
        The type of the variables:

            - c : continuous
            - u : unordered (discrete)
            - o : ordered (discrete)

        The string should contain a type specifier for each variable, so for
        example ``var_type='ccuo'``.
    bw: array_like or str, optional
        If an array, it is a fixed user-specified bandwidth.  If a string,
        should be one of:

            - normal_reference: normal reference rule of thumb (default)
            - cv_ml: cross validation maximum likelihood
            - cv_ls: cross validation least squares

    defaults: EstimatorSettings instance, optional
        The default values for (efficient) bandwidth estimation.

    Attributes
    ----------
    bw: array_like
        The bandwidth parameters.

    See Also
    --------
    KDEMultivariateConditional

    Examples
    --------
    >>> import statsmodels.api as sm
    >>> nobs = 300
    >>> np.random.seed(1234)  # Seed random generator
    >>> c1 = np.random.normal(size=(nobs,1))
    >>> c2 = np.random.normal(2, 1, size=(nobs,1))

    Estimate a bivariate distribution and display the bandwidth found:

    >>> dens_u = sm.nonparametric.KDEMultivariate(data=[c1,c2],
    ...     var_type='cc', bw='normal_reference')
    >>> dens_u.bw
    array([ 0.39967419,  0.38423292])
    """

    def __init__(self, data, var_type, bw=None, defaults=None):
        self.var_type = var_type
        self.k_vars = len(self.var_type)
        self.data = _adjust_shape(data, self.k_vars)
        self.data_type = var_type
        self.nobs, self.k_vars = np.shape(self.data)
        if self.nobs <= self.k_vars:
            raise ValueError("The number of observations must be larger "
                             "than the number of variables.")
        defaults = EstimatorSettings() if defaults is None else defaults
        self._set_defaults(defaults)
        if not self.efficient:
            self.bw = self._compute_bw(bw)
        else:
            self.bw = self._compute_efficient(bw)

    def __repr__(self):
        """Provide something sane to print."""
        rpr = "KDE instance\n"
        rpr += "Number of variables: k_vars = " + str(self.k_vars) + "\n"
        rpr += "Number of samples:   nobs = " + str(self.nobs) + "\n"
        rpr += "Variable types:      " + self.var_type + "\n"
        rpr += "BW selection method: " + self._bw_method + "\n"
        return rpr

    def pdf(self, data_predict=None):
        r"""
        Evaluate the probability density function.

        Parameters
        ----------
        data_predict: array_like, optional
            Points to evaluate at.  If unspecified, the training data is used.

        Returns
        -------
        pdf_est: array_like
            Probability density function evaluated at `data_predict`.

        Notes
        -----
        The probability density is given by the generalized product kernel
        estimator:

        .. math:: K_{h}(X_{i},X_{j}) =
            \prod_{s=1}^{q}h_{s}^{-1}k\left(\frac{X_{is}-X_{js}}{h_{s}}\right)
        """
        if data_predict is None:
            data_predict = self.data
        else:
            data_predict = _adjust_shape(data_predict, self.k_vars)

        pdf_est = []
        for i in range(np.shape(data_predict)[0]):
            pdf_est.append(gpke(self.bw, data=self.data,
                                data_predict=data_predict[i, :],
                                var_type=self.var_type) / self.nobs)

        pdf_est = np.squeeze(pdf_est)
        return pdf_est

    def _get_class_vars_type(self):
        """Helper method to be able to pass needed vars to _compute_subset."""
        class_type = 'KDEMultivariate'
        class_vars = (self.var_type, )
        return class_type, class_vars



class ContourIndexingError(BaseException):
    """Use when contour indices are confused"""
    pass


class IncompleteTDMSFileFormatError(BaseException):
    """Use for incomplete dataset (e.g. missing para.ini)"""


class InvalidTDMSFileFormatError(BaseException):
    """Use for invalid tdms files (e.g. unknown columns)"""
    pass


class InvalidVideoFileError(BaseException):
    """Used for bad video files"""
    pass



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
Computation of mean and standard deviation of grayscale values inside the
RT-DC event image mask.
"""
from __future__ import division, print_function, unicode_literals

import numpy as np


def get_bright(mask, image, ret_data="avg,sd"):
    """Compute avg and/or std of the event brightness

    The event brightness is defined by the gray-scale values of the
    image data within the event mask area.

    Parameters
    ----------
    mask: ndarray or list of ndarrays of shape (M,N) and dtype bool
        The mask values, True where the event is located in `image`.
    image: ndarray or list of ndarrays of shape (M,N)
        A 2D array that holds the image in form of grayscale values
        of an event.
    ret_data: str
        A comma-separated list of metrices to compute
        - "avg": compute the average
        - "sd": compute the standard deviation
        Selected metrics are returned in alphabetical order.

    Returns
    -------
    bright_avg: float or ndarray of size N
        Average image data within the contour
    bright_std: float or ndarray of size N
        Standard deviation of image data within the contour
    """
    # This method is based on a pull request by Maik Herbig.
    ret_avg = "avg" in ret_data
    ret_std = "sd" in ret_data

    if ret_avg + ret_std == 0:
        raise ValueError("No valid metrices selected!")

    if isinstance(mask, np.ndarray) and len(mask.shape) == 2:
        # We have a single image
        image = [image]
        mask = [mask]
        ret_list = False
    else:
        ret_list = True

    length = min(len(mask), len(image))

    # Results are stored in a separate array initialized with nans
    if ret_avg:
        avg = np.zeros(length, dtype=float) * np.nan
    if ret_std:
        std = np.zeros(length, dtype=float) * np.nan

    for ii in range(length):
        imgi = image[ii]
        mski = mask[ii]
        # Assign results
        if ret_avg:
            avg[ii] = np.mean(imgi[mski])
        if ret_std:
            std[ii] = np.std(imgi[mski])

    results = []
    # Keep alphabetical order
    if ret_avg:
        results.append(avg)
    if ret_std:
        results.append(std)

    if not ret_list:
        # Only return scalars
        results = [r[0] for r in results]

    if ret_avg+ret_std == 1:
        # Only return one column
        return results[0]

    return results



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""Utility methods"""
from __future__ import division, print_function, unicode_literals

import hashlib
import pathlib

import h5py
import numpy as np

from .compat import str_types


def hashfile(fname, blocksize=65536, count=0, hasher_class=hashlib.md5):
    """Compute md5 hex-hash of a file

    Parameters
    ----------
    fname: str
        path to the file
    blocksize: int
        block size in bytes read from the file
        (set to `0` to hash the entire file)
    count: int
        number of blocks read from the file
    """
    hasher = hasher_class()
    fname = pathlib.Path(fname)
    with fname.open('rb') as fd:
        buf = fd.read(blocksize)
        ii = 0
        while len(buf) > 0:
            hasher.update(buf)
            buf = fd.read(blocksize)
            ii += 1
            if count and ii == count:
                break
    return hasher.hexdigest()


def hashobj(obj):
    """Compute md5 hex-hash of a Python object"""
    return hashlib.md5(obj2str(obj)).hexdigest()


def obj2str(obj):
    """String representation of an object for hashing"""
    if isinstance(obj, str_types):
        return obj.encode("utf-8")
    elif isinstance(obj, pathlib.Path):
        return obj2str(str(obj))
    elif isinstance(obj, (bool, int, float)):
        return str(obj).encode("utf-8")
    elif obj is None:
        return b"none"
    elif isinstance(obj, np.ndarray):
        return obj.tobytes()
    elif isinstance(obj, tuple):
        return obj2str(list(obj))
    elif isinstance(obj, list):
        return b"".join(obj2str(o) for o in obj)
    elif isinstance(obj, dict):
        return obj2str(sorted(obj.items()))
    elif hasattr(obj, "identifier"):
        return obj2str(obj.identifier)
    elif isinstance(obj, h5py.Dataset):
        return obj2str(obj[0])
    else:
        raise ValueError("No rule to convert object '{}' to string.".
                         format(obj.__class__))



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""Viscosity computation for various media"""
from __future__ import division, print_function, unicode_literals

import warnings

import numpy as np

from ...warn import PipelineWarning


#: Media for which computation of viscosity is defined
KNOWN_MEDIA = ["CellCarrier", "CellCarrierB", "water"]


class TemperatureOutOfRangeWarning(PipelineWarning):
    pass


def get_viscosity(medium="CellCarrier", channel_width=20.0, flow_rate=0.16,
                  temperature=23.0):
    """Returns the viscosity for RT-DC-specific media

    Media that are not pure (e.g. ketchup or polymer solutions)
    often exhibit a non-linear relationship between shear rate
    (determined by the velocity profile) and shear stress
    (determined by pressure differences). If the shear stress
    grows non-linearly with the shear rate resulting in a slope
    in log-log space that is less than one, then we are talking about
    shear thinning. The viscosity is not a constant anymore (as it
    is e.g. for water). At higher flow rates, the viscosity becomes
    smaller, following a power law. Christoph Herold characterized
    shear thinning for the CellCarrier media :cite:`Herold2017`.
    The resulting formulae for computing the viscosities of these
    media at different channel widths, flow rates, and temperatures,
    are implemented here.

    Parameters
    ----------
    medium: str
        The medium to compute the viscosity for; Valid values
        are defined in :const:`KNOWN_MEDIA`.
    channel_width: float
        The channel width in µm
    flow_rate: float
        Flow rate in µL/s
    temperature: float or ndarray
        Temperature in °C

    Returns
    -------
    viscosity: float or ndarray
        Viscosity in mPa*s

    Notes
    -----
    - CellCarrier and CellCarrier B media are optimized for
      RT-DC measurements.
    - Values for the viscosity of water are computed using
      equation (15) from :cite:`Kestin_1978`.
    - A :class:`TemperatureOutOfRangeWarning` is issued if the
      input temperature range exceeds the temperature ranges given
      by :cite:`Herold2017` and :cite:`Kestin_1978`.
    """
    # also support lower-case media and a space before the "B"
    valmed = [v.lower() for v in KNOWN_MEDIA + ["CellCarrier B"]]
    medium = medium.lower()
    if medium not in valmed:
        raise ValueError("Invalid medium: {}".format(medium))

    # convert flow_rate from µL/s to m³/s
    # convert channel_width from µm to m
    term1 = 1.1856 * 6 * flow_rate * 1e-9 / (channel_width * 1e-6)**3 * 2 / 3

    if medium == "cellcarrier":
        temp_corr = (temperature / 23.2)**-0.866
        term2 = 0.6771 / 0.5928 + 0.2121 / (0.5928 * 0.677)
        eta = 0.179 * (term1 * term2)**(0.677 - 1) * temp_corr * 1e3
        if np.min(temperature) < 16 or np.max(temperature) > 26:
            # see figure (9) in Herold arXiv:1704.00572 (2017)
            warnings.warn("For CellCarrier, the temperature should be in "
                          + "[18, 26] degC! Got min/max of "
                          + "[{:.1f}, {:.1f}] degC.".format(
                              np.min(temperature), np.max(temperature)),
                          TemperatureOutOfRangeWarning)
    elif medium in ["cellcarrierb", "cellcarrier b"]:
        temp_corr = (temperature / 23.6)**-0.866
        term2 = 0.6771 / 0.5928 + 0.2121 / (0.5928 * 0.634)
        eta = 0.360 * (term1 * term2)**(0.634 - 1) * temp_corr * 1e3
        if np.min(temperature) < 16 or np.max(temperature) > 26:
            # see figure (9) in Herold arXiv:1704.00572 (2017)
            warnings.warn("For CellCarrier B, the temperature should be in "
                          + "[18, 26] degC! Got min/max of "
                          + "[{:.1f}, {:.1f}] degC.".format(
                              np.min(temperature), np.max(temperature)),
                          TemperatureOutOfRangeWarning)
    elif medium == "water":
        if np.min(temperature) < 0 or np.max(temperature) > 40:
            # see equation (15) in Kestin et al, J. Phys. Chem. 7(3) 1978
            warnings.warn("For water, the temperature should be in [0, 40] "
                          + "degC! Got min/max of "
                          + "[{:.1f}, {:.1f}] degC.".format(
                              np.min(temperature), np.max(temperature)),
                          TemperatureOutOfRangeWarning)
        eta0 = 1.002  # [mPa]
        right = (20-temperature) / (temperature + 96) \
            * (+ 1.2364
               - 1.37e-3 * (20 - temperature)
               + 5.7e-6 * (20 - temperature)**2
               )
        eta = eta0 * 10**right
    return eta



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals

import sys

import h5py
import numpy as np
import pytest

from dclab.rtdc_dataset import check, check_dataset, fmt_tdms, new_dataset, \
    write

from helper_methods import cleanup, example_data_dict, retrieve_data


@pytest.mark.filterwarnings(
    'ignore::dclab.rtdc_dataset.config.UnknownConfigurationKeyWarning')
def test_basic():
    h5path = retrieve_data("rtdc_data_hdf5_contour_image_trace.zip")
    viol, aler, info = check_dataset(h5path)
    # Features: Unknown key 'ncells'
    # Metadata: Mismatch [imaging] 'roi size x' and feature image (50 vs 90)
    # Metadata: Mismatch [imaging] 'roi size y' and feature image (90 vs 50)
    # Metadata: Missing key [fluorescence] 'channels installed'
    # Metadata: Missing key [fluorescence] 'laser count'
    # Metadata: Missing key [fluorescence] 'lasers installed'
    # Metadata: Missing key [fluorescence] 'samples per event'
    # Metadata: fluorescence channel count inconsistent
    assert len(viol) == 8
    # "HDF5: '/image': attribute 'CLASS' should be fixed-length ASCII string",
    # "HDF5: '/image': attribute 'IMAGE_SUBCLASS' should be fixed-length ...
    # "HDF5: '/image': attribute 'IMAGE_VERSION' should be fixed-length ...
    # "Metadata: Flow rates don't add up (sh 0.6 + sam 0.1 != channel 0.16)",
    # "Metadata: Flow rates don't add up (sh 0.6 + sam 0.1 != channel 0.16)",
    # "Metadata: Flow rates don't add up (sh 0.6 + sam 0.1 != channel 0.16)",
    # "Metadata: Missing key [fluorescence] 'channel 1 name'",
    # "Metadata: Missing key [fluorescence] 'channel 2 name'",
    # "Metadata: Missing key [fluorescence] 'channel 3 name'",
    # "Metadata: Missing key [setup] 'identifier'",
    # "Metadata: Missing section 'online_contour'"
    # "UnknownConfigurationKeyWarning: Unknown key 'exposure time' ...",
    # "UnknownConfigurationKeyWarning: Unknown key 'flash current' ...",
    assert len(aler) == 13
    assert "Data file format: hdf5" in info
    assert "Fluorescence: True" in info
    assert "Compression: Partial (1 of 25)" in info
    cleanup()


def test_complete():
    h5path = retrieve_data("rtdc_data_hdf5_rtfdc.zip")
    viol, aler, info = check_dataset(h5path)
    assert len(viol) == 0
    assert len(aler) == 0
    assert "Data file format: hdf5" in info
    assert "Fluorescence: True" in info
    cleanup()


def test_exact():
    h5path = retrieve_data("rtdc_data_traces_2flchan.zip")
    viol, aler, info = check_dataset(h5path)
    known_viol = [
        "Metadata: Missing key [fluorescence] 'channel count'",
        "Metadata: Missing key [fluorescence] 'channels installed'",
        "Metadata: Missing key [fluorescence] 'laser count'",
        "Metadata: Missing key [fluorescence] 'lasers installed'",
        "Metadata: Missing key [fluorescence] 'samples per event'",
    ]
    known_aler = [
        "Metadata: Missing key [fluorescence] 'channel 1 name'",
        "Metadata: Missing key [fluorescence] 'channel 2 name'",
        "Metadata: Missing key [online_contour] 'no absdiff'",
        "Metadata: Missing key [setup] 'identifier'",
        "Metadata: Missing key [setup] 'module composition'",
    ]
    known_info = [
        'Compression: None',
        'Data file format: tdms',
        'Fluorescence: True',
    ]
    assert set(viol) == set(known_viol)
    assert set(aler) == set(known_aler)
    assert set(info) == set(known_info)
    cleanup()


def test_icue():
    h5path = retrieve_data("rtdc_data_hdf5_rtfdc.zip")
    with check.IntegrityChecker(h5path) as ic:
        cues = ic.check()
    assert cues[0] != cues[1]
    levels = check.ICue.get_level_summary(cues)
    assert levels["info"] >= 2
    assert levels["alert"] == 0
    assert levels["violation"] == 0
    assert cues[0].msg in cues[0].__repr__()
    cleanup()


def test_ic_expand_section():
    ddict = example_data_dict(size=8472, keys=["area_um", "deform"])
    ds1 = new_dataset(ddict)
    ds2 = new_dataset(ddict)
    with check.IntegrityChecker(ds1) as ic:
        cues1 = ic.check_metadata_missing(expand_section=True)
    with check.IntegrityChecker(ds2) as ic:
        cues2 = ic.check_metadata_missing(expand_section=False)
    assert len(cues1) > len(cues2)


def test_ic_feature_size_scalar():
    ddict = example_data_dict(size=8472, keys=["area_um", "deform"])
    ddict["bright_sd"] = np.linspace(10, 20, 1000)
    ds = new_dataset(ddict)
    with check.IntegrityChecker(ds) as ic:
        cues = ic.check_feature_size()
    for cue in cues:
        if cue.category == "feature size":
            assert cue.msg == "Features: wrong event count: 'bright_sd' " \
                              + "(1000 of 8472)"
            break
    else:
        assert False


def test_ic_feature_size_trace():
    ddict = example_data_dict(size=8472, keys=["area_um", "deform"])
    ddict["trace"] = {"fl1_raw": [range(10)] * 1000}
    ds = new_dataset(ddict)
    with check.IntegrityChecker(ds) as ic:
        cues = ic.check_feature_size()
    for cue in cues:
        if cue.category == "feature size":
            assert cue.msg == "Features: wrong event count: 'trace/fl1_raw'" \
                              + " (1000 of 8472)"
            break
    else:
        assert False


def test_ic_fl_metadata_channel_names():
    ddict = example_data_dict(size=8472, keys=["area_um", "deform", "fl1_max"])
    ddict["trace"] = {"fl1_raw": [range(10)] * 1000}
    ds = new_dataset(ddict)
    with check.IntegrityChecker(ds) as ic:
        cues = ic.check_fl_metadata_channel_names()
    assert cues[0].category == "metadata missing"
    assert cues[0].cfg_section == "fluorescence"
    assert cues[0].cfg_key == "channel 1 name"


def test_ic_fl_metadata_channel_names_2():
    ddict = example_data_dict(size=8472, keys=["area_um", "deform"])
    ds = new_dataset(ddict)
    ds.config["fluorescence"]["channel 1 name"] = "peter"
    with check.IntegrityChecker(ds) as ic:
        cues = ic.check_fl_metadata_channel_names()
    assert cues[0].category == "metadata invalid"
    assert cues[0].cfg_section == "fluorescence"
    assert cues[0].cfg_key == "channel 1 name"


def test_ic_fl_num_channels():
    ddict = example_data_dict(size=8472, keys=["area_um", "deform"])
    ds = new_dataset(ddict)
    ds.config["fluorescence"]["channel count"] = 3
    ds.config["fluorescence"]["channel 1 name"] = "hans"
    with check.IntegrityChecker(ds) as ic:
        cues = ic.check_fl_num_channels()
    assert cues[0].category == "metadata wrong"
    assert cues[0].cfg_section == "fluorescence"
    assert cues[0].cfg_key == "channel count"


def test_ic_fl_num_lasers():
    ddict = example_data_dict(size=8472, keys=["area_um", "deform"])
    ds = new_dataset(ddict)
    ds.config["fluorescence"]["laser count"] = 3
    ds.config["fluorescence"]["laser 1 lambda"] = 550
    ds.config["fluorescence"]["laser 1 power"] = 20
    with check.IntegrityChecker(ds) as ic:
        cues = ic.check_fl_num_lasers()
    assert cues[0].category == "metadata wrong"
    assert cues[0].cfg_section == "fluorescence"
    assert cues[0].cfg_key == "laser count"


def test_ic_fmt_hdf5_image1():
    h5path = retrieve_data("rtdc_data_hdf5_rtfdc.zip")
    with h5py.File(h5path, "a") as h5:
        h5["events/image"].attrs["CLASS"] = "bad"
    with check.IntegrityChecker(h5path) as ic:
        cues = ic.check_fmt_hdf5()
    assert cues[0].category == "format HDF5"
    assert cues[0].msg == "HDF5: '/image': attribute 'CLASS' should be " \
                          + "fixed-length ASCII string"
    assert cues[0].level == "alert"
    cleanup()


@pytest.mark.skipif(sys.version_info < (3, 0),
                    reason="requires python3 or higher")
def test_ic_fmt_hdf5_image2():
    h5path = retrieve_data("rtdc_data_hdf5_rtfdc.zip")
    with h5py.File(h5path, "a") as h5:
        h5["events/image"].attrs["CLASS"] = np.string_("bad")
    with check.IntegrityChecker(h5path) as ic:
        cues = ic.check_fmt_hdf5()
    assert cues[0].category == "format HDF5"
    assert cues[0].msg == "HDF5: '/image': attribute 'CLASS' should have " \
                          + "value 'b'IMAGE''"
    assert cues[0].level == "alert"
    cleanup()


def test_ic_fmt_hdf5_image3():
    h5path = retrieve_data("rtdc_data_hdf5_rtfdc.zip")
    with h5py.File(h5path, "a") as h5:
        del h5["events/image"].attrs["CLASS"]
    with check.IntegrityChecker(h5path) as ic:
        cues = ic.check_fmt_hdf5()
    assert cues[0].category == "format HDF5"
    assert cues[0].msg == "HDF5: '/image': missing attribute 'CLASS'"
    assert cues[0].level == "alert"
    cleanup()


def test_ic_fmt_hdf5_logs():
    h5path = retrieve_data("rtdc_data_hdf5_rtfdc.zip")
    write(h5path, logs={
        "test": ["asdasd"*100],
        "M1_para.ini":  ["asdasd"*100],  # should be ignored
    }, mode="append")
    with check.IntegrityChecker(h5path) as ic:
        cues = ic.check_fmt_hdf5()
    assert len(cues) == 1
    assert cues[0].category == "format HDF5"
    assert cues[0].msg == 'Logs: test line 0 exceeds maximum line length 100'
    assert cues[0].level == "alert"
    cleanup()


def test_ic_metadata_bad():
    ddict = example_data_dict(size=8472, keys=["area_um", "deform"])
    ds = new_dataset(ddict)
    ds.config["experiment"]["run index"] = "1"
    with check.IntegrityChecker(ds) as ic:
        cues = ic.check_metadata_bad()
    assert cues[0].category == "metadata dtype"
    assert cues[0].level == "violation"
    assert cues[0].cfg_section == "experiment"
    assert cues[0].cfg_key == "run index"


def test_ic_metadata_choices():
    ddict = example_data_dict(size=8472, keys=["area_um", "deform"])
    ds = new_dataset(ddict)
    ds.config["setup"]["medium"] = "honey"
    with check.IntegrityChecker(ds) as ic:
        cues = ic.check_metadata_choices()
    assert cues[0].category == "metadata wrong"
    assert cues[0].level == "violation"
    assert cues[0].cfg_section == "setup"
    assert cues[0].cfg_key == "medium"


@pytest.mark.filterwarnings('ignore::dclab.rtdc_dataset.'
                            + 'ancillaries.ancillary_feature.'
                            + 'BadFeatureSizeWarning')
@pytest.mark.skipif(sys.version_info < (3, 6),
                    reason="requires python3.6 or higher")
def test_invalid_medium():
    h5path = retrieve_data("rtdc_data_minimal.zip")
    para = h5path.with_name("M1_para.ini")
    cfg = para.read_text().split("\n")
    cfg.insert(3, "Buffer Medium = unknown_bad!")
    para.write_text("\n".join(cfg))
    viol, _, _ = check_dataset(h5path)
    assert "Metadata: Invalid value [setup] medium: 'unknown_bad!'" in viol
    cleanup()


@pytest.mark.filterwarnings('ignore::dclab.rtdc_dataset.'
                            + 'ancillaries.ancillary_feature.'
                            + 'BadFeatureSizeWarning')
def test_load_with():
    h5path = retrieve_data("rtdc_data_minimal.zip")
    known_aler = [
        "Metadata: Missing key [setup] 'flow rate sample'",
        "Metadata: Missing key [setup] 'flow rate sheath'",
        "Metadata: Missing key [setup] 'identifier'",
        "Metadata: Missing key [setup] 'module composition'",
        "Metadata: Missing key [setup] 'software version'",
    ]
    known_viol = [
        "Features: wrong event count: 'contour' (14 of 156)",
        "Features: wrong event count: 'mask' (14 of 156)",
        "Metadata: Missing key [setup] 'medium'",
    ]
    with new_dataset(h5path) as ds:
        viol, aler, _ = check_dataset(ds)
        assert set(viol) == set(known_viol)
        assert set(aler) == set(known_aler)
    cleanup()


def test_missing_file():
    h5path = retrieve_data("rtdc_data_traces_2flchan.zip")
    h5path.with_name("M1_para.ini").unlink()
    try:
        check_dataset(h5path)
    except fmt_tdms.IncompleteTDMSFileFormatError:
        pass
    else:
        assert False
    cleanup()


def test_ml_class():
    """Test score data outside boundary"""
    data = {"ml_score_001": [.1, 10, -10, 0.01, .89],
            "ml_score_002": [.2, .1, .4, 0, .4],
            }
    ds = new_dataset(data)
    with check.IntegrityChecker(ds) as ic:
        cues = ic.check_ml_class()
        assert len(cues) == 1
        assert "ml_score_001" in cues[0].msg


@pytest.mark.filterwarnings('ignore::dclab.rtdc_dataset.'
                            + 'ancillaries.ancillary_feature.'
                            + 'BadFeatureSizeWarning')
def test_no_fluorescence():
    h5path = retrieve_data("rtdc_data_minimal.zip")
    _, _, info = check_dataset(h5path)
    known_info = [
        'Compression: None',
        'Data file format: tdms',
        'Fluorescence: False',
    ]
    assert set(info) == set(known_info)
    cleanup()


def test_temperature():
    # there are probably a million things wrong with this dataset, but
    # we are only looking for the temperature thing
    ddict = example_data_dict(size=8472, keys=["area_um", "deform", "temp"])
    ds = new_dataset(ddict)
    sstr = "Metadata: Missing key [setup] 'temperature', " \
           + "because the 'temp' feature is given"
    _, aler, _ = check_dataset(ds)
    assert sstr in aler


def test_wrong_samples_per_event():
    h5path = retrieve_data("rtdc_data_traces_2flchan.zip")
    with h5path.with_name("M1_para.ini").open("a") as fd:
        fd.write("Samples Per Event = 10\n")
    msg = "Metadata: wrong number of samples per event: fl1_median " \
          + "(expected 10, got 566)"
    viol, _, _ = check_dataset(h5path)
    assert msg in viol
    cleanup()


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" Test functions for loading contours
"""
from __future__ import print_function, unicode_literals

import os
import tempfile
import warnings

import numpy as np

from dclab.rtdc_dataset import new_dataset
import dclab.rtdc_dataset.config as dccfg

from helper_methods import retrieve_data, example_data_sets, cleanup


def equals(a, b):
    """Compare objects with allclose"""
    if isinstance(a, (dict, dccfg.Configuration, dccfg.ConfigurationDict)):
        for key in a:
            assert key in b, "key not in b"
            assert equals(a[key], b[key])
    elif isinstance(a, (float, int)):
        if np.isnan(a):
            assert np.isnan(b)
        else:
            assert np.allclose(a, b), "a={} vs b={}".format(a, b)
    else:
        assert a == b, "a={} vs b={}".format(a, b)
    return True


def test_config_basic():
    ds = new_dataset(retrieve_data(example_data_sets[1]))
    assert ds.config["imaging"]["roi size y"] == 96.
    cleanup()


def test_config_invalid_key():
    ds = new_dataset(retrieve_data(example_data_sets[1]))
    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        # Trigger a warning.
        ds.config["setup"]["invalid_key"] = "picard"
        # Verify some things
        assert len(w) == 1
        assert issubclass(w[-1].category, dccfg.UnknownConfigurationKeyWarning)
        assert "invalid_key" in str(w[-1].message)
    cleanup()


def test_config_save_load():
    # Download and extract data
    tdms_path = retrieve_data(example_data_sets[0])
    ds = new_dataset(tdms_path)
    cfg_file = tempfile.mktemp(prefix="test_dclab_rtdc_config_")
    ds.config.save(cfg_file)
    loaded = dccfg.Configuration(files=[cfg_file])
    assert equals(loaded, ds.config)
    cleanup()
    try:
        os.remove(cfg_file)
    except OSError:
        pass


def test_config_update():
    ds = new_dataset(retrieve_data(example_data_sets[1]))
    assert ds.config["imaging"]["roi size y"] == 96.
    ds.config["calculation"].update({
        "crosstalk fl12": 0.1,
        "crosstalk fl13": 0.2,
        "crosstalk fl21": 0.3,
        "crosstalk fl23": 0.4,
        "crosstalk fl31": 0.5,
        "crosstalk fl32": 0.6,
    })

    assert ds.config["calculation"]["crosstalk fl12"] == 0.1
    assert ds.config["calculation"]["crosstalk fl13"] == 0.2
    assert ds.config["calculation"]["crosstalk fl21"] == 0.3
    assert ds.config["calculation"]["crosstalk fl23"] == 0.4
    assert ds.config["calculation"]["crosstalk fl31"] == 0.5
    assert ds.config["calculation"]["crosstalk fl32"] == 0.6
    cleanup()


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



cdef unsigned char point_in_polygon(Py_ssize_t nr_verts, double *xp, double *yp,
                                    double x, double y) nogil

cdef void points_in_polygon(Py_ssize_t nr_verts, double *xp, double *yp,
                            Py_ssize_t nr_points, double *x, double *y,
                            unsigned char *result) nogil



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function

import sys
import time

import numpy as np
import pytest

import dclab
from dclab import new_dataset
from dclab.features.contour import get_contour, get_contour_lazily
from dclab.features.volume import get_volume

from helper_methods import retrieve_data, cleanup


def test_artefact():
    ds = new_dataset(retrieve_data("rtdc_data_hdf5_mask_artefact.zip"))
    # This would raise a "dclab.features.contour.NoValidContourFoundError:
    # Event 1, No contour found!" in dclab version <= 0.22.1
    cont = ds["contour"][1]
    assert len(cont) == 37, "just to be sure there really is something"
    cleanup()


def test_lazy_contour_basic():
    ds = new_dataset(retrieve_data("rtdc_data_hdf5_mask_contour.zip"))
    masks = ds["mask"][:]
    cont1 = get_contour_lazily(masks)
    cont2 = get_contour(masks)
    for ii in range(len(ds)):
        assert np.all(cont1[ii] == cont2[ii])
    cleanup()


@pytest.mark.skipif(sys.version_info < (3, 3),
                    reason="perf_counter requires python3.3 or higher")
def test_lazy_contour_timing():
    ds = new_dataset(retrieve_data("rtdc_data_hdf5_mask_contour.zip"))
    masks = ds["mask"][:]
    t0 = time.perf_counter()
    get_contour_lazily(masks)
    t1 = time.perf_counter()
    get_contour(masks)
    t2 = time.perf_counter()
    assert t2-t1 > 10*(t1-t0)
    cleanup()


def test_lazy_contour_type():
    ds1 = new_dataset(retrieve_data("rtdc_data_hdf5_mask_contour.zip"))
    c1 = ds1["contour"]
    # force computation of contour data
    ds1._events._features.remove("contour")
    c2 = ds1["contour"]
    assert isinstance(c1, dclab.rtdc_dataset.fmt_hdf5.H5ContourEvent)
    assert isinstance(c2, dclab.features.contour.LazyContourList)
    cleanup()


def test_simple_contour():
    ds = new_dataset(retrieve_data("rtdc_data_traces_video_bright.zip"))
    # Note: contour "3" in ds is bad!
    cin = ds["contour"][2]
    mask = np.zeros_like(ds["image"][2], dtype="bool")
    mask[cin[:, 1], cin[:, 0]] = True
    cout = get_contour(mask)
    # length
    assert len(cin) == len(cout)
    # simple presence test
    for ci in cin:
        assert ci in cout
    # order
    for ii in range(1, len(cin)):
        c2 = np.roll(cin, ii, axis=0)
        if np.all(c2 == cout):
            break
    else:
        assert False, "contours not matching, check orientation?"
    cleanup()


def test_volume():
    ds = new_dataset(retrieve_data("rtdc_data_hdf5_mask_contour.zip"))
    mask = [mi for mi in ds["mask"]]
    cont1 = [ci for ci in ds["contour"]]
    cont2 = get_contour(mask)

    kw = dict(pos_x=ds["pos_x"],
              pos_y=ds["pos_y"],
              pix=ds.config["imaging"]["pixel size"])

    v1 = get_volume(cont=cont1, **kw)
    v2 = get_volume(cont=cont2, **kw)

    assert np.allclose(v1, v2)
    cleanup()


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function

import numpy as np

import dclab

from helper_methods import example_data_dict


def test_downsample_index():
    keys = ["area_um", "deform", "time", "frame", "fl3_width"]
    ddict = example_data_dict(size=8472, keys=keys)
    ds = dclab.new_dataset(ddict)

    ds.apply_filter()
    x, y, index = ds.get_downsampled_scatter(xax="area_um",
                                             yax="deform",
                                             downsample=100,
                                             ret_mask=True)
    assert np.all(x == ds["area_um"][index])
    assert np.all(y == ds["deform"][index])

    # also with log scale
    x2, y2, index2 = ds.get_downsampled_scatter(xax="area_um",
                                                yax="deform",
                                                downsample=100,
                                                xscale="log",
                                                ret_mask=True)
    assert np.all(x2 == ds["area_um"][index2])
    assert np.all(y2 == ds["deform"][index2])


def test_downsample_log():
    keys = ["area_um", "deform", "time", "frame", "fl3_width"]
    ddict = example_data_dict(size=8472, keys=keys)
    ds = dclab.new_dataset(ddict)

    ds.apply_filter()
    xlin, ylin = ds.get_downsampled_scatter(downsample=100,
                                            xscale="linear",
                                            yscale="linear")
    xlog, ylog = ds.get_downsampled_scatter(downsample=100,
                                            xscale="log",
                                            yscale="log")
    assert not np.all(xlin == xlog)
    assert not np.all(ylin == ylog)
    for x in xlog:
        assert x in ds["area_um"]
    for y in ylog:
        assert y in ds["deform"]


def test_downsample_log_invalid():
    ddict = {"area_um": np.linspace(-2, 10, 100),
             "deform": np.linspace(.1, .5, 100)}
    ds = dclab.new_dataset(ddict)
    ds.apply_filter()
    xlog, ylog = ds.get_downsampled_scatter(downsample=99,
                                            xscale="log",
                                            yscale="log",
                                            remove_invalid=False,
                                            )
    for x in xlog:
        assert x in ds["area_um"]
    for y in ylog:
        assert y in ds["deform"]

    assert xlog.size == 99
    assert ylog.size == 99
    assert xlog.min() < 0


def test_downsample_log_invalid_index():
    ddict = {"area_um": np.linspace(-2, 10, 100),
             "deform": np.linspace(.1, .5, 100)}
    ds = dclab.new_dataset(ddict)
    ds.apply_filter()
    x, y, index = ds.get_downsampled_scatter(xax="area_um",
                                             yax="deform",
                                             downsample=99,
                                             xscale="log",
                                             yscale="log",
                                             remove_invalid=False,
                                             ret_mask=True
                                             )
    assert np.all(x == ds["area_um"][index])
    assert np.all(y == ds["deform"][index])


def test_downsample_log_invalid_removed():
    ddict = {"area_um": np.array([-100, 0, 100, 200, 300, 400, 500]),
             "deform": np.array([.1, .2, .3, .4, .5, 6, np.nan])}
    ds = dclab.new_dataset(ddict)
    ds.apply_filter()
    xlog, ylog = ds.get_downsampled_scatter(downsample=6,
                                            xscale="log",
                                            yscale="log",
                                            remove_invalid=True,
                                            )
    assert xlog.size == 4
    assert ylog.size == 4
    assert np.all(xlog == np.array([100, 200, 300, 400]))
    assert np.all(ylog == np.array([.3, .4, .5, 6]))


def test_downsample_log_invalid_removed_index():
    ddict = {"area_um": np.array([-100, 0, 100, 200, 300, 400, 500]),
             "deform": np.array([.1, .2, .3, .4, .5, 6, np.nan])}
    ds = dclab.new_dataset(ddict)
    ds.apply_filter()
    x, y, index = ds.get_downsampled_scatter(xax="area_um",
                                             yax="deform",
                                             downsample=6,
                                             xscale="log",
                                             yscale="log",
                                             remove_invalid=True,
                                             ret_mask=True
                                             )
    assert np.all(x == ds["area_um"][index])
    assert np.all(y == ds["deform"][index])


def test_downsample_none():
    keys = ["area_um", "deform", "time", "frame", "fl3_width"]
    ddict = example_data_dict(size=8472, keys=keys)
    ds = dclab.new_dataset(ddict)

    ds.apply_filter()
    _, _, idx = ds.get_downsampled_scatter(downsample=0, ret_mask=True)
    assert np.sum(idx) == 8472


def test_downsample_none2():
    keys = ["area_um", "deform", "time", "frame", "fl3_width"]
    ddict = example_data_dict(size=8472, keys=keys)
    ds = dclab.new_dataset(ddict)

    filtflt = {"enable filters": False}

    cfg = {"filtering": filtflt}
    ds.config.update(cfg)
    ds.apply_filter()
    _, _, idx = ds.get_downsampled_scatter(downsample=100, ret_mask=True)

    assert np.sum(idx) == 100
    assert np.sum(ds.filter.all) == 8472

    filtflt["enable filters"] = True
    ds.config.update(cfg)
    ds.apply_filter()
    _, _, idx = ds.get_downsampled_scatter(downsample=100, ret_mask=True)

    assert np.sum(idx) == 100
    assert np.sum(ds.filter.all) == 8472


def test_downsample_up():
    """
    Likely causes removal of too many points and requires
    re-inserting them.
    """
    keys = ["area_um", "deform", "time", "frame", "fl3_width"]
    ddict = example_data_dict(size=10000, keys=keys)
    ds = dclab.new_dataset(ddict)

    ds.apply_filter()
    _, _, idx = ds.get_downsampled_scatter(downsample=9999, ret_mask=True)
    assert np.sum(idx) == 9999
    ds.get_downsampled_scatter(downsample=9999)


def test_downsample_yes():
    """Simple downsampling test"""
    keys = ["area_um", "deform", "time", "frame", "fl3_width"]
    ddict = example_data_dict(size=8472, keys=keys)
    ds = dclab.new_dataset(ddict)

    ds.apply_filter()
    _, _, idx = ds.get_downsampled_scatter(downsample=100, ret_mask=True)
    assert np.sum(idx) == 100
    ds.get_downsampled_scatter(downsample=100)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function

import numpy as np

from dclab import new_dataset
from dclab.features.bright import get_bright

from helper_methods import retrieve_data, cleanup


def test_simple_bright():
    ds = new_dataset(retrieve_data("rtdc_data_traces_video_bright.zip"))
    for ii in range(2, 7):
        # This stripped dataset has only 7 video frames / contours
        image = ds["image"][ii]
        mask = ds["mask"][ii]
        avg, std = get_bright(mask=mask, image=image, ret_data="avg,sd")
        assert np.allclose(avg, ds["bright_avg"][ii])
        assert np.allclose(std, ds["bright_sd"][ii])
    cleanup()


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""Check RT-DC datasets for completeness"""
from __future__ import unicode_literals

import copy
import functools
import warnings

import h5py
import numpy as np

from .core import RTDCBase
from .load import load_file

from .. import definitions as dfn


#: log names that end with these strings are not checked
IGNORED_LOG_NAMES = {
    "_para.ini",
    "_image.ini",
    "FG_Config.mcf",
    "parameters.txt",
    "_SoftwareSettings.ini",
    "dckit-history",
}

#: keys that must be present for every measurement
IMPORTANT_KEYS = {
    "experiment": [
        "date",
        "event count",
        "run index",
        "sample",
        "time"],
    "imaging": [
        "flash device",
        "flash duration",
        "frame rate",
        "pixel size",
        "roi position x",
        "roi position y",
        "roi size x",
        "roi size y"],
    "setup": [
        "channel width",
        "chip region",
        "flow rate",
        "medium"],
}

#: keys that must be present for fluorescence measurements
IMPORTANT_KEYS_FL = {
    "fluorescence": [
        "bit depth",
        "channel count",
        "channels installed",
        "laser count",
        "lasers installed",
        "sample rate",
        "samples per event",
        "signal max",
        "signal min",
        "trace median"],
}

#: maximum line length in log files
LOG_MAX_LINE_LENGTH = 100

#: keys that are optional
OPTIONAL_KEYS = {
    "setup": [
        "temperature"],
}

#: valid metadata choices
VALID_CHOICES = {
    "setup": {
        "medium": ["CellCarrier", "CellCarrierB", "water", "other"],
    }
}


@functools.total_ordering
class ICue(object):
    def __init__(self, msg, level, category, data=None,
                 cfg_section=None, cfg_key=None, cfg_choices=None):
        """Integrity cue"""
        #: human readable message
        self.msg = msg
        #: machine-readable data associated with the check
        self.data = data
        #: severity level ("violation", "alert", or "info")
        self.level = level
        #: fail category
        self.category = category
        #: section (only for categories "missing metadata", "bad metadata")
        self.cfg_section = cfg_section
        #: key (only for categories "missing metadata", "bad metadata");
        #: can be omitted e.g. to communicate that the entire section is
        #: missing
        self.cfg_key = cfg_key
        #: allowed choices for the specific [section]: key combination
        #: (only for categories "missing metadata", "bad metadata")
        self.cfg_choices = cfg_choices
        if self.cfg_choices is None:
            if (cfg_section in VALID_CHOICES
                    and cfg_key in VALID_CHOICES[cfg_section]):
                self.cfg_choices = VALID_CHOICES[cfg_section][cfg_key]

    def __eq__(self, other):
        leveld = {"info": 0,
                  "violation": 1,
                  "alert": 2,
                  }
        return ((leveld[self.level], self.cfg_section or "",
                 self.cfg_key or "", self.category, self.msg) ==
                (leveld[other.level], other.cfg_section or "",
                 other.cfg_key or "", other.category, other.msg))

    def __lt__(self, other):
        leveld = {"info": 0,
                  "violation": 1,
                  "alert": 2, }
        return ((leveld[self.level], self.cfg_section or "",
                 self.cfg_key or "", self.category, self.msg) <
                (leveld[other.level], other.cfg_section or "",
                 other.cfg_key or "", other.category, other.msg))

    def __repr__(self):
        return "<ICue: '{}' at 0x{}>".format(self.msg, hex(id(self)))

    @staticmethod
    def get_level_summary(cues):
        """For a list of ICue, return the abundance of all levels"""
        levels = {"info": 0,
                  "alert": 0,
                  "violation": 0}
        for cue in cues:
            levels[cue.level] += 1
        return levels


class IntegrityChecker(object):
    def __init__(self, path_or_ds):
        """Check the integrity of a dataset

        The argument must be either a path to an .rtdc or .tdms file
        or an instance of `RTDCBase`. If a path is given, then all
        warnings (e.g. UnknownConfigurationKeyWarning) are catched
        and added to the cue list.

        Usage:

        .. code:: python

            ic = IntegrityChecker("/path/to/data.rtdc")
            cues = ic.check()

        """
        self.warn_cues = []
        if isinstance(path_or_ds, RTDCBase):
            self.ds = path_or_ds
            self.finally_close = False
        else:
            with warnings.catch_warnings(record=True) as ws:
                warnings.simplefilter("always")
                self.ds = load_file(path_or_ds)
                for ww in ws:
                    self.warn_cues.append(ICue(
                        msg="{}: {}".format(ww.category.__name__, ww.message),
                        level="alert",
                        category="warning"))
            self.finally_close = True

    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        # close the file
        if self.finally_close and hasattr(self.ds, "__exit__"):
            self.ds.__exit__(type, value, tb)

    @property
    def has_fluorescence(self):
        if ("fluorescence" in self.ds
            or "fl1_max" in self.ds
            or "fl2_max" in self.ds
                or "fl3_max" in self.ds):
            fl = True
        else:
            fl = False
        return fl

    def check(self, **kwargs):
        """Run all checks

        This method calls all class methods that start with `check_`.
        `kwargs` are passed to all methods. Possible options are:

        - "expand_section" `bool`: add a cue for every missing
          metadata key if a section is missing
        """
        cues = []
        funcs = IntegrityChecker.__dict__
        for ff in sorted(funcs.keys()):
            if (ff.startswith("check_fl_") and not self.has_fluorescence):
                # skip
                continue
            elif ff.startswith("check_"):
                cues += funcs[ff](self, **kwargs)
        return sorted(self.warn_cues + cues)

    def check_compression(self, **kwargs):
        cues = []
        if self.ds.format == "tdms":
            compression = "None"
            data = {"compressed": 0,
                    "total": 1,
                    "uncompressed": 0,
                    }
        elif self.ds.format == "hdf5":
            def iter_count_compression(h5):
                comp = 0
                noco = 0
                for key in h5:
                    obj = h5[key]
                    if isinstance(obj, h5py.Dataset):
                        if obj.compression is None:
                            noco += 1
                        else:
                            comp += 1
                    elif isinstance(obj, h5py.Group):
                        coi, noi = iter_count_compression(obj)
                        comp += coi
                        noco += noi
                    else:
                        raise ValueError("Unknown object: {}".format(obj))
                return comp, noco
            comp, noco = iter_count_compression(self.ds._h5)
            if noco == 0:
                compression = "All"
            elif comp == 0:
                compression = "None"
            else:
                compression = "Partial ({} of {})".format(comp, noco+comp)
            data = {"compressed": comp,
                    "total": noco+comp,
                    "uncompressed": noco,
                    }
        else:
            compression = "Unknown"
            data = None
        cues.append(ICue(
            msg="Compression: {}".format(compression),
            level="info",
            category="general",
            data=data))
        return cues

    def check_feat_index(self, **kwargs):
        """Up until"""
        cues = []
        lends = len(self.ds)
        if "index" in self.ds:
            if not np.all(self.ds["index"] == np.arange(1, lends+1)):
                cues.append(ICue(
                    msg="The index feature is not enumerated correctly",
                    level="violation",
                    category="feature data"))
        return cues

    def check_feature_size(self, **kwargs):
        cues = []
        lends = len(self.ds)
        for feat in self.ds.features_innate:
            msg = "Features: wrong event count: '{}' ({} of {})"
            if feat == "trace":
                for tr in list(self.ds["trace"].keys()):
                    if len(self.ds["trace"][tr]) != lends:
                        cues.append(ICue(
                            msg=msg.format("trace/" + tr,
                                           len(self.ds["trace"][tr]),
                                           lends),
                            level="violation",
                            category="feature size"))
            else:
                if len(self.ds[feat]) != lends:
                    cues.append(ICue(
                        msg=msg.format(feat, len(self.ds[feat]), lends),
                        level="violation",
                        category="feature size"))
        return cues

    def check_features_unknown_hdf5(self, **kwargs):
        """check for feature column names"""
        cues = []
        if self.ds.format == "hdf5":
            for feat in self.ds._h5["events"]:
                if not dfn.feature_exists(feat):
                    cues.append(ICue(
                        msg="Features: Unknown key '{}'".format(feat),
                        level="violation",
                        category="feature unknown"))
        return cues

    def check_fl_metadata_channel_names(self, **kwargs):
        cues = []
        for ii in range(1, 4):
            chn = "channel {} name".format(ii)
            fli = "fl{}_max".format(ii)
            if (fli in self.ds
                    and chn not in self.ds.config["fluorescence"]):
                # Channel names must be defined when there is
                # a corresponding fluorescence signal.
                cues.append(ICue(
                    msg="Metadata: Missing key [{}] '{}'".format(
                        "fluorescence", chn),
                    level="alert",
                    category="metadata missing",
                    cfg_section="fluorescence",
                    cfg_key=chn))
            elif (fli not in self.ds
                  and chn in self.ds.config["fluorescence"]):
                # Channel names must not be defined when there is
                # no corresponding fluorescence signal.
                cues.append(ICue(
                    msg="Metadata: Unused key defined [{}] '{}'".format(
                        "fluorescence", chn),
                    level="alert",
                    category="metadata invalid",
                    cfg_section="fluorescence",
                    cfg_key=chn))
        return cues

    def check_fl_num_channels(self, **kwargs):
        cues = []
        # check for number of channels
        if "channel count" in self.ds.config["fluorescence"]:
            chc1 = self.ds.config["fluorescence"]["channel count"]
            chc2 = 0
            for ii in range(1, 4):
                chn = "channel {} name".format(ii)
                ecn = "fl{}_max".format(ii)
                if (chn in self.ds.config["fluorescence"] and
                        ecn in self.ds._events):
                    chc2 += 1
            if chc1 != chc2:
                cues.append(ICue(
                    msg="Metadata: fluorescence channel count inconsistent",
                    level="violation",
                    category="metadata wrong",
                    cfg_section="fluorescence",
                    cfg_key="channel count"))
        return cues

    def check_fl_num_lasers(self, **kwargs):
        cues = []
        # check for number of lasers
        if "laser count" in self.ds.config["fluorescence"]:
            lsc1 = self.ds.config["fluorescence"]["laser count"]
            lsc2 = 0
            for ii in range(1, 4):
                kl = "laser {} lambda".format(ii)
                kp = "laser {} power".format(ii)
                if (kl in self.ds.config["fluorescence"] and
                        kp in self.ds.config["fluorescence"] and
                        self.ds.config["fluorescence"][kp] != 0):
                    lsc2 += 1
            if lsc1 != lsc2:
                cues.append(ICue(
                    msg="Metadata: fluorescence laser count inconsistent",
                    level="violation",
                    category="metadata wrong",
                    cfg_section="fluorescence",
                    cfg_key="laser count"))
        return cues

    def check_fl_samples_per_event(self, **kwargs):
        cues = []
        # check for samples per event
        if "samples per event" in self.ds.config["fluorescence"]:
            spe = self.ds.config["fluorescence"]["samples per event"]
            if "trace" in self.ds:
                for key in self.ds["trace"].keys():
                    spek = self.ds["trace"][key][0].size
                    if spek != spe:
                        cues.append(ICue(
                            msg="Metadata: wrong number of samples per "
                                + "event: {} (expected {}, got {})".format(
                                    key, spe, spek),
                            level="violation",
                            category="metadata wrong",
                            cfg_section="fluorescence",
                            cfg_key="samples per event"))
        return cues

    def check_flow_rate(self, **kwargs):
        """Make sure sheath and sample flow rates add up"""
        cues = []
        if ("setup" in self.ds.config
            and "flow rate" in self.ds.config["setup"]
            and "flow rate sample" in self.ds.config["setup"]
                and "flow rate sheath" in self.ds.config["setup"]):
            frsum = self.ds.config["setup"]["flow rate"]
            frsam = self.ds.config["setup"]["flow rate sample"]
            frshe = self.ds.config["setup"]["flow rate sheath"]
            if not np.allclose(frsum, frsam+frshe):
                for k in ["flow rate", "flow rate sheath", "flow rate sample"]:
                    cues.append(ICue(
                        msg="Metadata: Flow rates don't add up (sheath "
                            + "{:g} + sample {:g} != channel {:g})".format(
                                frshe, frsam, frsum),
                        level="alert",
                        category="metadata wrong",
                        cfg_section="setup",
                        cfg_key=k))
        return cues

    def check_fmt_hdf5(self, **kwargs):
        cues = []
        # hdf5-based checks
        if self.ds.format == "hdf5":
            # check meta data of images
            for feat in ["image", "mask"]:
                if feat in self.ds._events:
                    imdat = self.ds[feat]
                    for key, val in [['CLASS', b'IMAGE'],
                                     ['IMAGE_VERSION', b'1.2'],
                                     ['IMAGE_SUBCLASS', b'IMAGE_GRAYSCALE']]:
                        if key not in imdat.attrs:
                            cues.append(ICue(
                                msg="HDF5: '/{}': missing ".format(feat)
                                    + "attribute '{}'".format(key),
                                level="alert",
                                category="format HDF5"))
                        elif imdat.attrs.get_id(key).dtype.char != "S":
                            cues.append(ICue(
                                msg="HDF5: '/{}': attribute '{}' ".format(
                                    feat, key)
                                + "should be fixed-length ASCII string",
                                level="alert",
                                category="format HDF5"))
                        elif imdat.attrs[key] != val:
                            cues.append(ICue(
                                msg="HDF5: '/{}': attribute '{}' ".format(
                                    feat, key)
                                + "should have value '{}'".format(val),
                                level="alert",
                                category="format HDF5"))
            # check length of logs
            with h5py.File(self.ds.path, mode="r") as h5:
                if "logs" in h5:
                    logs = h5["logs"]
                    for logname in logs.keys():
                        # ignore tmds meta data log files
                        lign = [logname.endswith(n) for n in IGNORED_LOG_NAMES]
                        if sum(lign):
                            continue
                        log = logs[logname]
                        for ii in range(len(log)):
                            if len(log[ii]) > LOG_MAX_LINE_LENGTH:
                                cues.append(ICue(
                                    msg="Logs: {} line {} ".format(logname, ii)
                                            + "exceeds maximum line length "
                                            + "{}".format(LOG_MAX_LINE_LENGTH),
                                            level="alert",
                                            category="format HDF5"))
        return cues

    def check_info(self, **kwargs):
        cues = []
        cues.append(ICue(
            msg="Fluorescence: {}".format(self.has_fluorescence),
            level="info",
            category="general"))
        cues.append(ICue(
            msg="Data file format: {}".format(self.ds.format),
            level="info",
            category="general"))
        return cues

    def check_metadata_bad(self, **kwargs):
        cues = []
        # check for meta data types
        for sec in self.ds.config:
            for key in self.ds.config[sec]:
                if sec in dfn.CFG_ANALYSIS:
                    # TODO:
                    # - properly test against analysis keywords
                    #   (filtering, calculation)
                    pass
                elif not isinstance(self.ds.config[sec][key],
                                    dfn.config_types[sec][key]):
                    cues.append(ICue(
                        msg="Metadata: Datatype of [{}] '{}'".format(sec, key)
                            + "must be '{}'".format(
                                dfn.config_types[sec][key]),
                        level="violation",
                        category="metadata dtype",
                        cfg_section=sec,
                        cfg_key=key))
        # check for ROI size
        if ("imaging" in self.ds.config
            and "roi size x" in self.ds.config["imaging"]
                and "roi size y" in self.ds.config["imaging"]):
            for ii, roi in enumerate(["roi size y", "roi size x"]):
                for feat in ["image", "mask"]:
                    if feat in self.ds:
                        soll = self.ds[feat][0].shape[ii]
                        ist = self.ds.config["imaging"][roi]
                        if soll != ist:
                            cues.append(ICue(
                                msg="Metadata: Mismatch [imaging] "
                                        + "'{}' and feature {} ".format(roi,
                                                                        feat)
                                        + "({} vs {})".format(ist, soll),
                                        level="violation",
                                        category="metadata wrong",
                                        cfg_section="imaging",
                                        cfg_key=roi))
        return cues

    def check_metadata_choices(self, **kwargs):
        cues = []
        for sec in VALID_CHOICES:
            for key in VALID_CHOICES[sec]:
                if sec in self.ds.config and key in self.ds.config[sec]:
                    val = self.ds.config[sec][key]
                    if val not in VALID_CHOICES[sec][key]:
                        cues.append(ICue(
                            msg="Metadata: Invalid value [{}] {}: '{}'".format(
                                sec, key, val),
                            level="violation",
                            category="metadata wrong",
                            cfg_section=sec,
                            cfg_key=key,
                            cfg_choices=VALID_CHOICES[sec][key]))
        return cues

    def check_metadata_missing(self, expand_section=True, **kwargs):
        cues = []
        # These "must" be present:
        tocheck = copy.deepcopy(IMPORTANT_KEYS)
        if self.has_fluorescence:
            tocheck.update(IMPORTANT_KEYS_FL)
        # These sections "should" be fully present (except if in OPTIONAL_KEYS)
        tocheck_sec_aler = ["experiment", "imaging", "online_contour", "setup"]

        for sec in tocheck:
            if sec not in self.ds.config and not expand_section:
                cues.append(ICue(
                    msg="Metadata: Missing section '{}'".format(sec),
                    level="violation",
                    category="metadata missing",
                    cfg_section=sec))
            else:
                for key in dfn.config_keys[sec]:
                    if (key in tocheck[sec]
                            and key not in self.ds.config[sec]):
                        cues.append(ICue(
                            msg="Metadata: Missing key [{}] '{}'".format(sec,
                                                                         key),
                            level="violation",
                            category="metadata missing",
                            cfg_section=sec,
                            cfg_key=key))
                    elif ((sec in tocheck_sec_aler
                           and key not in self.ds.config[sec])
                          and not (sec in OPTIONAL_KEYS
                                   and key in OPTIONAL_KEYS[sec])):
                        # The section should be present, but the key is not.
                        # The key is not in the optional key list.
                        # Note that fluorescence data is not checked here
                        cues.append(ICue(
                            msg="Metadata: Missing key [{}] '{}'".format(sec,
                                                                         key),
                            level="alert",
                            category="metadata missing",
                            cfg_section=sec,
                            cfg_key=key))
        # search again (soft)
        for sec in tocheck_sec_aler:
            if sec in tocheck:
                # already treated above (hard)
                continue
            if sec not in self.ds.config and not expand_section:
                cues.append(ICue(
                    msg="Metadata: Missing section '{}'".format(sec),
                    level="alert",
                    category="metadata missing",
                    cfg_section=sec,
                    cfg_key=key))
            else:
                for key in dfn.config_keys[sec]:
                    if key not in self.ds.config[sec]:
                        cues.append(ICue(
                            msg="Metadata: Missing key [{}] '{}'".format(sec,
                                                                         key),
                            level="alert",
                            category="metadata missing",
                            cfg_section=sec,
                            cfg_key=key))
        # check for temperature
        if "temp" in self.ds:
            if "temperature" not in self.ds.config["setup"]:
                cues.append(ICue(
                    msg="Metadata: Missing key [setup] 'temperature', "
                        + "because the 'temp' feature is given",
                    level="alert",
                    category="metadata missing",
                    cfg_section="setup",
                    cfg_key="temperature"))
        return cues

    def check_ml_class(self, **kwargs):
        """Try to comput ml_class feature and display error message"""
        cues = []
        if "ml_class" in self.ds:
            try:
                self.ds["ml_class"]
            except ValueError as e:
                cues.append(ICue(
                    msg=e.args[0],
                    level="violation",
                    category="feature data"))
        return cues


def check_dataset(path_or_ds):
    """Check whether a dataset is complete

    Parameters
    ----------
    path_or_ds: str or RTDCBase
        Full path to a dataset on disk or an instance of RTDCBase

    Returns
    -------
    violations: list of str
        Dataset format violations (hard)
    alerts: list of str
        Dataset format alerts (soft)
    info: list of str
        Dataset information
    """
    aler = []
    info = []
    viol = []
    with IntegrityChecker(path_or_ds) as ic:
        # perform all checks
        icues = ic.check(expand_section=False)
        for cue in icues:
            if cue.level == "info":
                info.append(cue.msg)
            elif cue.level == "alert":
                aler.append(cue.msg)
            elif cue.level == "violation":
                viol.append(cue.msg)
    return sorted(viol), sorted(aler), sorted(info)



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""DCOR client interface"""
from __future__ import division, print_function, unicode_literals

import numpy as np

from ..compat import PyImportError, lru_cache
from .. import definitions as dfn
from ..util import hashobj

from .config import Configuration
from .core import RTDCBase

try:
    import requests
except PyImportError:
    REQUESTS_AVAILABLE = False
else:
    REQUESTS_AVAILABLE = True


class DCORAccessError(BaseException):
    pass


class APIHandler(object):
    """Handles the DCOR api with caching for simple queries"""
    #: these are cached to minimize network usage
    cache_queries = ["metadata", "size", "feature_list", "valid"]
    #: DCOR API Keys in the current session
    api_keys = []

    def __init__(self, url, api_key=""):
        self.url = url
        self.api_key = api_key
        self._cache = {}

    @classmethod
    def add_api_key(cls, api_key):
        """Add an API Key to the base class

        When accessing the DCOR API, all available API Keys are
        used to access a resource (trial and error).
        """
        if api_key.strip() and api_key not in APIHandler.api_keys:
            APIHandler.api_keys.append(api_key)

    def _get(self, query, feat=None, trace=None, event=None, api_key=""):
        qstr = "&query={}".format(query)
        if feat is not None:
            qstr += "&feature={}".format(feat)
        if trace is not None:
            qstr += "&trace={}".format(trace)
        if event is not None:
            qstr += "&event={}".format(event)
        apicall = self.url + qstr
        req = requests.get(apicall, headers={"Authorization": api_key})
        return req.json()

    def get(self, query, feat=None, trace=None, event=None):
        if query in APIHandler.cache_queries and query in self._cache:
            result = self._cache[query]
        else:
            for api_key in [self.api_key] + APIHandler.api_keys:
                req = self._get(query, feat, trace, event, api_key)
                if req["success"]:
                    self.api_key = api_key  # remember working key
                    break
            else:
                raise DCORAccessError("Cannot access {}: {}".format(
                    query, req["error"]["message"]))
            result = req["result"]
            if query in APIHandler.cache_queries:
                self._cache[query] = result
        return result


class EventFeature(object):
    """Helper class for accessing non-scalar features event-wise"""

    def __init__(self, feature, api):
        self.identifier = api.url + ":" + feature  # for caching ancillaries
        self.feature = feature
        self.api = api

    def __iter__(self):
        for idx in range(len(self)):
            yield self[idx]

    @lru_cache(maxsize=100)
    def __getitem__(self, event):
        data = self.api.get(query="feature", feat=self.feature, event=event)
        return np.asarray(data)

    def __len__(self):
        return int(self.api.get(query="size"))


class EventTrace(object):
    """Helper class for accessing traces event-wise"""

    def __init__(self, trace, api):
        self.identifier = api.url + ":" + trace  # for caching ancillaries
        self.trace = trace
        self.api = api

    def __iter__(self):
        for idx in range(len(self)):
            yield self[idx]

    @lru_cache(maxsize=100)
    def __getitem__(self, event):
        data = self.api.get(query="trace", trace=self.trace, event=event)
        return np.asarray(data)

    def __len__(self):
        return int(self.api.get(query="size"))


class EventTraceFeature(object):
    """Helper class for accessing traces"""

    def __init__(self, api):
        self.identifier = api.url + ":traces"
        self.api = api
        self.traces = api.get(query="trace_list")
        self._trace_objects = {}

    def __contains__(self, key):
        return key in self.traces

    def __getitem__(self, trace):
        if trace in self.traces:
            if trace not in self._trace_objects:
                self._trace_objects[trace] = EventTrace(trace, self.api)
            return self._trace_objects[trace]
        else:
            raise KeyError("trace '{}' not found!".format(trace))

    def keys(self):
        return self.traces


class FeatureCache(object):
    """Download and cache (scalar only) features from DCOR"""

    def __init__(self, api):
        self.api = api
        self._features = self.api.get(query="feature_list")
        self._scalar_cache = {}
        self._nonsc_features = {}

    def __contains__(self, key):
        return key in self._features

    def __getitem__(self, key):
        # user-level checking is done in core.py
        assert dfn.feature_exists(key)
        if key not in self._features:
            raise KeyError("Feature '{}' not found!".format(key))

        if key in self._scalar_cache:
            return self._scalar_cache[key]
        elif dfn.scalar_feature_exists(key):
            # download the feature and cache it
            feat = np.asarray(self.api.get(query="feature", feat=key))
            self._scalar_cache[key] = feat
            return feat
        elif key == "trace":
            if "trace" not in self._nonsc_features:
                self._nonsc_features["trace"] = EventTraceFeature(self.api)
            return self._nonsc_features["trace"]
        else:
            if key not in self._nonsc_features:
                self._nonsc_features[key] = EventFeature(key, self.api)
            return self._nonsc_features[key]

    def __iter__(self):
        # dict-like behavior
        for key in self.keys():
            yield key

    def keys(self):
        return self._features


class RTDC_DCOR(RTDCBase):
    def __init__(self, url, use_ssl=None, host="dcor.mpl.mpg.de",
                 api_key="", *args, **kwargs):
        """Wrap around the DCOR API

        Parameters
        ----------
        url: str
            Full URL or resource identifier; valid values are

            - `<https://dcor.mpl.mpg.de/api/3/action/dcserv?id=
              b1404eb5-f661-4920-be79-5ff4e85915d5>`_
            - dcor.mpl.mpg.de/api/3/action/dcserv?id=b1404eb5-f
              661-4920-be79-5ff4e85915d5
            - b1404eb5-f661-4920-be79-5ff4e85915d5
        use_ssl: bool
            Set this to False to disable SSL (should only be used for
            testing). Defaults to None (does not force SSL if the URL
            starts with "http://").
        host: str
            The host machine (used if the host is not given in `url`)
        api_key: str
            API key to access private resources
        *args:
            Arguments for `RTDCBase`
        **kwargs:
            Keyword arguments for `RTDCBase`

        Attributes
        ----------
        path: str
            Full URL to the DCOR resource
        """
        if not REQUESTS_AVAILABLE:
            raise PyImportError("Package `requests` required for DCOR format!")

        super(RTDC_DCOR, self).__init__(*args, **kwargs)

        self._hash = None
        self.path = RTDC_DCOR.get_full_url(url, use_ssl, host)
        self.api = APIHandler(url=self.path, api_key=api_key)

        # Parse configuration
        self.config = Configuration(cfg=self.api.get(query="metadata"))

        # Get size
        self._size = int(self.api.get(query="size"))

        # Setup events
        self._events = FeatureCache(self.api)

        # Override logs property with HDF5 data
        self.logs = {}

        self.title = "{} - M{}".format(self.config["experiment"]["sample"],
                                       self.config["experiment"]["run index"])

        # Set up filtering
        self._init_filters()

    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        pass

    def __len__(self):
        return self._size

    @staticmethod
    def get_full_url(url, use_ssl, host):
        """Return the full URL to a DCOR resource

        Parameters
        ----------
        url: str
            Full URL or resource identifier; valid values are

            - https://dcor.mpl.mpg.de/api/3/action/dcserv?id=caab96f6-
              df12-4299-aa2e-089e390aafd5'
            - dcor.mpl.mpg.de/api/3/action/dcserv?id=caab96f6-df12-
              4299-aa2e-089e390aafd5
            - caab96f6-df12-4299-aa2e-089e390aafd5
        use_ssl: bool
            Set this to False to disable SSL (should only be used for
            testing). Defaults to None (does not force SSL if the URL
            starts with "http://").
        host: str
            Use this host if it is not specified in `url`
        """
        if use_ssl is None:
            if url.startswith("http://"):
                # user wanted it that way
                web = "http"
            else:
                web = "https"
        elif use_ssl:
            web = "https"
        else:
            web = "http"
        if url.count("://"):
            base = url.split("://", 1)[1]
        else:
            base = url
        if base.count("/"):
            host, api = base.split("/", 1)
        else:
            api = "api/3/action/dcserv?id=" + base
        new_url = "{}://{}/{}".format(web, host, api)
        return new_url

    @property
    def hash(self):
        """Hash value based on file name and content"""
        if self._hash is None:
            tohash = [self.path]
            self._hash = hashobj(tohash)
        return self._hash



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""Class for on-the-fly conversion of contours to masks"""
from __future__ import division, print_function, unicode_literals

import numpy as np
import scipy.ndimage as ndi


class MaskColumn(object):
    def __init__(self, rtdc_dataset):
        """Computes mask from contour data"""
        self.contour = rtdc_dataset["contour"]
        self.image = rtdc_dataset["image"]
        self.identifier = self.contour.identifier
        self.config = rtdc_dataset.config
        self._shape = None

    def __getitem__(self, idx):
        mask = np.zeros(self._img_shape, dtype=bool)
        conti = self.contour[idx]
        mask[conti[:, 1], conti[:, 0]] = True
        ndi.morphology.binary_fill_holes(mask, output=mask)
        return mask

    def __len__(self):
        if self._img_shape != (0, 0):
            lc = len(self.contour)
        else:
            lc = 0
        return lc

    @property
    def _img_shape(self):
        if self._shape is None:
            cfgim = self.config["imaging"]
            if self.image:
                # get shape from image column
                self._shape = self.image.shape
            elif "roi size x" in cfgim and "roi size y" in cfgim:
                # get shape from config (this is less reliable than getting
                # the shape from the image; there were measurements with
                # wrong config keys)
                self._shape = (cfgim["roi size y"], cfgim["roi size x"])
            else:
                # no shape available
                self._shape = (0, 0)
        return self._shape



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""RT-DC hdf5 format"""
from __future__ import division, print_function, unicode_literals

from distutils.version import LooseVersion
import pathlib

import h5py
import numpy as np

from .. import definitions as dfn
from ..util import hashobj, hashfile

from .config import Configuration
from .core import RTDCBase


#: rtdc files exported with dclab prior to this version are not supported
MIN_DCLAB_EXPORT_VERSION = "0.3.3.dev2"

#: dictionary of defective features, defined by HDF5 attributes;
#: if a value matches the given HDF5 attribute, the feature is
#: considered defective
DEFECTIVE_FEATURES = {
    # feature: [HDF5_attribute, matching_value]
    # In Shape-In 2.0.6, there was a wrong variable cast for the
    # feature "aspect".
    "aspect": [["setup:software version", "ShapeIn 2.0.6"],
               ["setup:software version", "ShapeIn 2.0.7"],
               ]
}


class OldFormatNotSupportedError(BaseException):
    pass


class UnknownKeyWarning(UserWarning):
    pass


class H5ContourEvent(object):
    def __init__(self, h5group):
        self.h5group = h5group
        self.identifier = h5group["0"][:]

    def __getitem__(self, key):
        return self.h5group[str(key)][:]

    def __iter__(self):
        for idx in range(len(self)):
            yield self[idx]

    def __len__(self):
        return len(self.h5group)


class H5Events(object):
    def __init__(self, h5):
        self._h5 = h5
        self._features = sorted(self._h5["events"].keys())
        # make sure that "trace" is not empty
        if "trace" in self._features and len(self._h5["events"]["trace"]) == 0:
            self._features.remove("trace")

    def __contains__(self, key):
        if self._is_defective_feature(key):
            contained = False
        else:
            contained = key in self._features
        return contained

    def __getitem__(self, key):
        # user-level checking is done in core.py
        assert dfn.feature_exists(key), "Feature '{}' not valid!".format(key)
        data = self._h5["events"][key]
        if key in ["image", "trace"]:
            return data
        elif key == "contour":
            return H5ContourEvent(data)
        elif key == "mask":
            return H5MaskEvent(data)
        else:
            return data[:]

    def __iter__(self):
        # dict-like behavior
        for key in self.keys():
            yield key

    def _is_defective_feature(self, key):
        """Whether or not the stored feature is defective"""
        defective = False
        if key in self._features:
            # feature exists in the HDF5 file
            if key in DEFECTIVE_FEATURES:
                # workaround machinery for sorting out defective features
                for attr, value in DEFECTIVE_FEATURES[key]:
                    if attr in self._h5.attrs:
                        defective = self._h5.attrs[attr] == value
                    if defective:
                        break
        return defective

    def keys(self):
        features = []
        for key in self._features:
            # check for defective features
            if not self._is_defective_feature(key):
                features.append(key)
        return features


class H5Logs(object):
    def __init__(self, h5):
        self._h5 = h5

    def __getitem__(self, key):
        if "logs" in self._h5:
            log = list(self._h5["logs"][key])
            if isinstance(log[0], bytes):
                log = [li.decode("utf") for li in log]
        else:
            raise KeyError("No logs in {}!".format(self._h5.file.filename))
        return log

    def __iter__(self):
        # dict-like behavior
        for key in self.keys():
            yield key

    def __len__(self):
        return len(self.keys())

    def keys(self):
        if "logs" in self._h5:
            names = sorted(self._h5["logs"].keys())
        else:
            names = []
        return names


class H5MaskEvent(object):
    """Cast uint8 masks to boolean"""

    def __init__(self, h5group):
        self.h5group = h5group
        # identifier required because "mask" is used for computation
        # of ancillary feature "contour".
        self.identifier = str(self.h5group.parent.parent.file)

    def __getitem__(self, idx):
        return np.asarray(self.h5group[idx], dtype=bool)

    def __iter__(self):
        for idx in range(len(self)):
            yield self[idx]

    def __len__(self):
        return len(self.h5group)

    @property
    def attrs(self):
        return self.h5group.attrs

    @property
    def shape(self):
        return self.h5group.shape


class RTDC_HDF5(RTDCBase):
    def __init__(self, h5path, *args, **kwargs):
        """HDF5 file format for RT-DC measurements

        Parameters
        ----------
        h5path: str or pathlib.Path
            Path to a '.tdms' measurement file.
        *args:
            Arguments for `RTDCBase`
        **kwargs:
            Keyword arguments for `RTDCBase`

        Attributes
        ----------
        path: pathlib.Path
            Path to the experimental HDF5 (.rtdc) file
        """
        super(RTDC_HDF5, self).__init__(*args, **kwargs)

        h5path = pathlib.Path(h5path)
        self._hash = None
        self.path = h5path

        # Setup events
        self._h5 = h5py.File(h5path, mode="r")
        self._events = H5Events(self._h5)

        # Parse configuration
        self.config = RTDC_HDF5.parse_config(h5path)

        # Override logs property with HDF5 data
        self.logs = H5Logs(self._h5)

        # check version
        rtdc_soft = self.config["setup"]["software version"]
        if rtdc_soft.startswith("dclab "):
            rtdc_ver = LooseVersion(rtdc_soft.split(" ")[1])
            if rtdc_ver < LooseVersion(MIN_DCLAB_EXPORT_VERSION):
                msg = "The file {} was created ".format(self.path) \
                      + "with dclab {} which is ".format(rtdc_ver) \
                      + "not supported anymore! Please rerun " \
                      + "dclab-tdms2rtdc / export the data again."
                raise OldFormatNotSupportedError(msg)

        self.title = "{} - M{}".format(self.config["experiment"]["sample"],
                                       self.config["experiment"]["run index"])

        # Set up filtering
        self._init_filters()

    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        # close the HDF5 file
        self._h5.close()

    @staticmethod
    def can_open(h5path):
        """Check whether a given file is in the .rtdc file format"""
        h5path = pathlib.Path(h5path)
        if h5path.suffix == ".rtdc":
            return True
        else:
            # we don't know the extension; check for the "events" group
            canopen = False
            try:
                # This is a workaround for Python2 where h5py cannot handle
                # unicode file names.
                with h5path.open("rb") as fd:
                    h5 = h5py.File(fd, "r")
                    if "events" in h5:
                        canopen = True
            except IOError:
                # not an HDF5 file
                pass
            return canopen

    @staticmethod
    def parse_config(h5path):
        """Parse the RT-DC configuration of an hdf5 file"""
        with h5py.File(h5path, mode="r") as fh5:
            h5attrs = dict(fh5.attrs)

        # Convert byte strings to unicode strings
        # https://github.com/h5py/h5py/issues/379
        for key in h5attrs:
            if isinstance(h5attrs[key], bytes):
                h5attrs[key] = h5attrs[key].decode("utf-8")

        config = Configuration()
        for key in h5attrs:
            section, pname = key.split(":")
            if pname not in dfn.config_funcs[section]:
                # Add the value as a string (this will issue
                # a UnknownConfigurationKeyWarning in config.py)
                config[section][pname] = h5attrs[key]
            else:
                typ = dfn.config_funcs[section][pname]
                config[section][pname] = typ(h5attrs[key])
        return config

    @property
    def hash(self):
        """Hash value based on file name and content"""
        if self._hash is None:
            tohash = [self.path.name]
            # Hash a maximum of ~1MB of the hdf5 file
            tohash.append(hashfile(self.path, blocksize=65536, count=20))
            self._hash = hashobj(tohash)
        return self._hash



#cython: cdivision=True
#cython: boundscheck=False
#cython: nonecheck=False
#cython: wraparound=False


cdef inline unsigned char point_in_polygon(Py_ssize_t nr_verts, double *xp,
                                           double *yp, double x,
                                           double y) nogil:
    """Test whether point lies inside a polygon.

    Parameters
    ----------
    nr_verts : int
        Number of vertices of polygon.
    xp, yp : double array
        Coordinates of polygon with length nr_verts.
    x, y : double
        Coordinates of point.
    """
    cdef Py_ssize_t i
    cdef unsigned char c = 0
    cdef Py_ssize_t j = nr_verts - 1
    for i in range(nr_verts):
        if (
            (((yp[i] <= y) and (y < yp[j])) or
            ((yp[j] <= y) and (y < yp[i])))
            and (x < (xp[j] - xp[i]) * (y - yp[i]) / (yp[j] - yp[i]) + xp[i])
        ):
            c = not c
        j = i
    return c


cdef void points_in_polygon(Py_ssize_t nr_verts, double *xp, double *yp,
                            Py_ssize_t nr_points, double *x, double *y,
                            unsigned char *result) nogil:
    """Test whether points lie inside a polygon.

    Parameters
    ----------
    nr_verts : int
        Number of vertices of polygon.
    xp, yp : double array
        Coordinates of polygon with length nr_verts.
    nr_points : int
        Number of points to test.
    x, y : double array
        Coordinates of points.
    result : unsigned char array
        Test results for each point.
    """
    cdef Py_ssize_t n
    for n in range(nr_points):
        result[n] = point_in_polygon(nr_verts, xp, yp, x[n], y[n])



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
This library contains classes and methods for the analysis
of real-time deformability cytometry (RT-DC) datasets.
"""
from __future__ import division, print_function, unicode_literals

from . import definitions as dfn  # noqa: F401
from . import features  # noqa: F401
from . import isoelastics  # noqa: F401
from . import kde_contours  # noqa: F401
from . import kde_methods  # noqa: F401
from .polygon_filter import PolygonFilter  # noqa: F401
from . import rtdc_dataset  # noqa: F401
from .rtdc_dataset import new_dataset  # noqa: F401
from . import statistics  # noqa: F401

from ._version import version as __version__  # noqa: F401



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
Class for efficiently handling image/video data
"""
from __future__ import division, print_function, unicode_literals

import pathlib
import sys
import warnings

import imageio
import numpy as np

from .exc import InvalidVideoFileError


ISWIN = sys.platform.startswith("win")


class CorruptFrameWarning(UserWarning):
    pass


class InitialFrameMissingWarning(CorruptFrameWarning):
    pass


class SlowVideoWarning(UserWarning):
    pass


class ImageColumn(object):
    def __init__(self, rtdc_dataset):
        """A wrapper for ImageMap that takes into account event offsets

        Event offsets appear when the first event that is recorded in the
        tdms files does not have a corresponding cell image in the video
        file.
        """
        fname = self.find_video_file(rtdc_dataset)
        self.identifier = fname
        if fname is not None:
            self._image_data = ImageMap(fname)
        else:
            self._image_data = []
        conf = rtdc_dataset.config
        self.event_offset = int(conf["fmt_tdms"]["video frame offset"])
        self.video_file = fname
        self._shape = None

    def __getitem__(self, idx):
        idnew = int(idx-self.event_offset)
        if idnew < 0:
            # No data - show a dummy image instead
            warnings.warn("Frame {} in {} ".format(idnew, self.identifier)
                          + "is not defined; replacing with dummy image!",
                          InitialFrameMissingWarning)
            cdata = self.dummy
        else:
            if hasattr(imageio.plugins.ffmpeg, "CannotReadFrameError"):
                # imageio<2.5.0
                excs = IndexError, imageio.plugins.ffmpeg.CannotReadFrameError
            else:
                # imageio>=2.5.0
                excs = IndexError
            try:
                cdata = self._image_data[idnew]
            except excs:
                # The avi is corrupt. Return a dummy image.
                warnings.warn("Frame {} in {} ".format(idnew, self.identifier)
                              + "is corrupt; replacing with dummy image!",
                              CorruptFrameWarning)
                cdata = self.dummy
        return cdata

    def __len__(self):
        length = len(self._image_data)
        if length:
            length = length + self.event_offset
        return length

    @property
    def dummy(self):
        """Returns a dummy image"""
        cdata = np.zeros(self.shape, dtype=np.uint8)
        return cdata

    @property
    def shape(self):
        if self._shape is None:
            f0 = self._image_data[0].shape
            self._shape = (f0[0], f0[1])
        return self._shape

    @staticmethod
    def find_video_file(rtdc_dataset):
        """Tries to find a video file that belongs to an RTDC dataset

        Returns None if no video file is found.
        """
        video = None
        if rtdc_dataset._fdir.exists():
            # Cell images (video)
            videos = [v.name for v in rtdc_dataset._fdir.glob("*.avi")]
            # Filter videos according to measurement number
            meas_id = rtdc_dataset._mid
            videos = [v for v in videos if v.split("_")[0] == meas_id]
            videos.sort()
            if len(videos) != 0:
                # Defaults to first avi file
                video = videos[0]
                # g/q video file names. q comes first.
                for v in videos:
                    if v.endswith("imag.avi"):
                        video = v
                        break
                    # add this here, because fRT-DC measurements also contain
                    # videos ..._proc.avi
                    elif v.endswith("imaq.avi"):
                        video = v
                        break
        if video is None:
            return None
        else:
            vpath = rtdc_dataset._fdir / video
            if vpath.stat().st_size < 64:
                # the video file is empty
                raise InvalidVideoFileError("Bad video '{}'!".format(vpath))
            else:
                return vpath


class ImageMap(object):
    def __init__(self, fname):
        """Access a video file of an RT-DC dataset

        Initialize this class with a video file.
        """
        fname = pathlib.Path(fname)
        self._length = None
        # video handle:
        self._cap = None
        # filename
        if not fname.exists():
            raise OSError("file does not exist: {}".format(fname))
        self.filename = fname

    def __del__(self):
        if self._cap is not None:
            if ISWIN and self._cap._proc is not None:
                # This is a workaround for windows when pytest fails due
                # to "OSError: [WinError 6] The handle is invalid",
                # which is somehow related to the fact that "_proc.kill()"
                # must be called twice (in "close()" and in this case) in
                # order to terminate the process and due to the fact the
                # we are not using the with-statement in combination
                # with imageio.get_reader().
                self._cap._proc.kill()
            self._cap.close()

    def __getitem__(self, idx):
        """Returns the requested frame from the video in gray scale"""
        cap = self.video_handle
        try:
            cellimg = cap.get_data(idx)
        except IndexError:
            self._get_image_workaround_seek(idx)
        else:
            if np.all(cellimg == 0):
                cellimg = self._get_image_workaround_seek(idx)
        # Convert to grayscale
        if len(cellimg.shape) == 3:
            cellimg = np.array(cellimg[:, :, 0])
        return cellimg

    def __len__(self):
        """Returns the length of the video or `True` if the length cannot be
        determined.
        """
        if self._length is None:
            cap = self.video_handle
            if hasattr(cap, "count_frames"):  # imageio>=2.5.0
                length = cap.count_frames()
            else:  # imageio<2.5.0
                try:
                    length = len(cap)
                except TypeError:
                    # length is set to inf (it would generally still be
                    # possible to rescue a limted number of frames, but
                    # other parts of the dataset are probably also broken
                    # (aborted measurement), so for the sake of simplicity
                    # we stop here)
                    raise InvalidVideoFileError(
                        "Video file has unknown length '{}'!".format(
                            self.filename))
            self._length = length
        return self._length

    def _get_image_workaround_seek(self, idx):
        """Same as __getitem__ but seek through the video beforehand

        This is a workaround for an all-zero image returned by `imageio`.
        """
        cap = self.video_handle
        mult = 50
        # get the first frame to be on the safe side
        cap.get_data(101)  # frame above 100 (don't know why)
        cap.get_data(0)  # seek to zero
        for ii in range(idx//mult):
            cap.get_data(ii*mult)
        final = cap.get_data(idx)
        if not np.all(final == 0):
            # This means we succeeded
            warnings.warn("Seeking video file does not work, used workaround "
                          + "which is slow!", SlowVideoWarning)
        return final

    @property
    def video_handle(self):
        if self._cap is None:
            try:
                self._cap = imageio.get_reader(self.filename)
            except OSError:
                raise InvalidVideoFileError(
                    "Broken video file '{}'!".format(self.filename))
        return self._cap
